"""A module for post processing BigDFT calculations.

"""


def _system_command(command, options):
    """
    Run the command as ``os.system (command + options)``

    Args:
       command (str): the actual command to run.
       options (str): the options to pass to the command.
    """
    from subprocess import call

    command_str = command + " " + options
    call(command_str, shell=True)


def _get_datadir(log):
    """
    Given a log file, this returns the path to the data directory.

    Args:
      log (Logfile): logfile from a BigDFT run.
    Returns:
      str: path to the data directory.
    """
    from os.path import join
    return join(log.srcdir, log.data_directory)


class BigDFTool(object):
    """

    This module defines the actions that can be performed using the
    ``bigdft-tool`` python script. Such a script sometimes invokes the
    ``memguess`` executable, which is executed serially. Sometimes the
    ``utilities`` main program is necessary to take care of more time
    consuming operations which require parallelism. We might also use
    the ``obabel`` wrapper provided by BigDFT.

    Args:
      omp (int): number of OpenMP threads.
        It defaults to the $OMP_NUM_THREADS variable in the environment,
        if present, otherwise it fixes the run to 1 thread.
      mpi_run (str): define the MPI command to be used.
        It defaults to the value $BIGDFT_MPIRUN of the environment, if present.
        When using this calculator into a job submission script, the value of
        $BIGDFT_MPIRUN variable may be set appropriately to launch the bigdft
        executable.
    """

    def __init__(self, omp=1, mpi_run=None):
        from futile.YamlArgparse import get_python_function
        from futile import YamlIO
        from os import environ
        from os.path import join
        from functools import partial
        from copy import deepcopy

        # Executables
        self.bigdftroot = environ['BIGDFT_ROOT']
        # environ['OMP_NUM_THREADS'] = str(omp)
        if not mpi_run:
            mpi_run = environ["BIGDFT_MPIRUN"]

        # Load the dictionary that defines all the operations.
        # This path should be easier to specify if you use the python
        # egg setup.
        with open(join(environ['PYTHONPATH'], "BigDFT", "Database",
                       "postprocess.yaml")) as ifile:
            db = YamlIO.load(stream=ifile)[0]

        # Create the subroutines of the class from the dictionary
        for action, spec in db.items():
            naction = action.replace("_", "-")
            nspec = deepcopy(spec)
            cmd = join(self.bigdftroot, nspec["tool"])
            if nspec["mpi"]:
                cmd = mpi_run + " " + cmd
            else:
                nspec["args"]["mpirun"] = {"default": "" + mpi_run + ""}
            nspec["args"]["action"] = {"default": "" + naction + ""}
            my_action = get_python_function(partial(self._invoke_command, cmd),
                                            action, nspec)
            setattr(self, action, my_action)

    def _auto_fragment_distance(self, system, log, cutoff, verbose, rand,
                                kxs, pv, frag_indices):
        from copy import deepcopy
        from random import choice

        msys = deepcopy(system)

        if verbose:
            print("Entering main loop")
        for i in range(0, len(system) - 1):
            if verbose and len(system) > 20:
                step = int(len(system) / 20.0)
                if i % step == 0:
                    print(str(5 * (i / step)) + "% done", len(msys))

            # Find the least pure fragment to modify
            target = min(pv, key=pv.get)
            target_purity = pv[target]
            if abs(target_purity) < cutoff:
                break

            if rand:
                target_list = [pid for pid in pv if abs(pv[pid]) > cutoff]
                target = choice(target_list)

            # Find the neighbor to add in
            addition = msys.get_nearest_fragment(target)

            # Combine
            msys[target] = msys[target] + deepcopy(msys[addition])
            frag_indices[target] = frag_indices[target] + \
                deepcopy(frag_indices[addition])
            del msys[addition]
            del pv[addition]
            del frag_indices[addition]

            self.run_compute_purity(msys, log, kxs=kxs, frag_list=[
                                    target], frag_indices=frag_indices)
            pv[target] = msys[target].purity_indicator

        return msys

    def _auto_fragment_bondorder(self, system, log, cutoff, verbose, rand,
                                 kxs, pv, frag_indices):
        from copy import deepcopy
        from random import choice
        from numpy import infty

        msys = deepcopy(system)

        # Cutoff for how far to look for a fragment to add.
        distcut = 10.0

        if verbose:
            print("Entering main loop")
        for i in range(0, len(system) - 1):
            if verbose and len(system) > 20:
                step = int(len(system) / 20.0)
                if i % step == 0:
                    print(str(5 * (i / step)) + "% done", len(msys))

            # Find the least pure fragment to modify
            target = min(pv, key=pv.get)
            target_purity = pv[target]
            if abs(target_purity) < cutoff:
                break

            if rand:
                target_list = [pid for pid in pv if abs(pv[pid]) > cutoff]
                target = choice(target_list)

            # Find some nearby neighbors to potentially add
            k = 3*len(msys[target])
            addition_list = msys.get_k_nearest_fragments(target, k=k,
                                                         cutoff=distcut)

            # Find the neighbor to add in
            bo = self.fragment_bond_order(msys, [target], addition_list, log,
                                          kxs=kxs, frag_indices=frag_indices)
            bo = bo[target]

            # In case nothing was found, we widen the search distance.
            if len(bo) == 0:
                distcut = infty
                continue

            addition = max(bo, key=bo.get)

            # Combine
            msys[target] = msys[target] + deepcopy(msys[addition])
            frag_indices[target] = frag_indices[target] + \
                deepcopy(frag_indices[addition])
            del msys[addition]
            del pv[addition]
            del frag_indices[addition]

            self.run_compute_purity(msys, log, kxs=kxs, frag_list=[
                                    target], frag_indices=frag_indices)
            pv[target] = msys[target].purity_indicator

        return msys

    def auto_fragment(self, system, log, cutoff, verbose=False, rand=False,
                      refine=False, kxs=None, criteria="distance"):
        """
        Refragments a system based on the results of a previous calculation.

        The auto fragment protocol searches for the least pure fragment, and
        then merges it with its nearest neighbor. This is repeated until the
        purity values of the entire system are less than the provided cutoff.
        You can also set the rand parameter to have the protocol merge random
        non pure fragments instead.

        The refine option performs a refinement step after the fragmentation
        is complete. For each pure fragment, we break it up again, and then
        use the random merge approach within that fragment.

        Args:
          system (System): instance of a System class, which defines the
            fragments we will use.
          log (Logfile): instance of a Logfile class
          cutoff (float): the termination criteria. When the worst fragment is
            more pure than this cutoff, the calculation stops.
          verbose (bool): whether to print out as it proceeds.
          rand (bool): whether to randomly select fragments to merge or not.
          refine (bool): whether to do the additional refinement step.
          kxs (scipy.sparse.csr_matrix): the density times the overlap,
            if it has been precomputed.
          criteria (string): either distance or bondorder.

        Returns:
          (System): a refragmented system where each fragment has a purity
          value better than the cutoff.
        """
        from copy import deepcopy
        from BigDFT.Fragments import System, Fragment

        msys = deepcopy(system)
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        frag_indices = self.get_frag_indices(msys, log)
        self.run_compute_purity(msys, log, kxs=kxs, frag_indices=frag_indices)
        pv = {k: v.purity_indicator for k, v in msys.items()}

        if criteria == "distance":
            msys = self._auto_fragment_distance(msys, log, cutoff, verbose,
                                                rand, kxs, pv, frag_indices)
        elif criteria == "bondorder":
            msys = self._auto_fragment_bondorder(msys, log, cutoff, verbose,
                                                 rand, kxs, pv, frag_indices)
        else:
            raise ValueError("Criteria must be distance or bondorder")

        # Give the fragments a name
        msys = msys.rename_fragments()

        # Refinement Loop
        if refine:
            if verbose:
                print("Entering refine loop")
            msys2 = System()
            for count, fragid in enumerate(msys):
                if verbose and len(msys) > 20:
                    step = int(len(msys) / 20.0)
                    if count % step == 0:
                        print(str(5 * (count / step)) + "% done", len(msys))
                # Split up the subsystem
                subsys = System()
                for i, at in enumerate(msys[fragid]):
                    subsys["ATOM:" + str(i)] = Fragment([at])

                # Autofragment
                msubsys = self.auto_fragment(subsys, log, cutoff, rand=True,
                                             kxs=kxs)

                # Add those fragments to the main system
                for sfrag in msubsys.values():
                    sname = "FRAG:" + str(len(msys2) + 1)
                    msys2[sname] = sfrag

            msys = msys2.rename_fragments()

        return msys

    def set_fragment_connectivity(self, frag, log, kxs=None):
        """
        Compute the connectivty of a given fragment using the fragment bond
        order, and sets that information to frag.conmat.

        Args:
          frag (BigDFT.Fragments.Fragment): the fragment to compute the
            connectivity of.
          log (Logfile): the Logfile of a calculation of the full system.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
        """
        from BigDFT.Fragments import System
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        # First convert to an atomic system
        atomsys = System()
        for i, at in enumerate(frag):
            atomsys["ATOM:"+str(i)] = [at]

        # Compute the bond order.
        bo = self.fragment_bond_order(atomsys, atomsys.keys(), atomsys.keys(),
                                      log, kxs=kxs)

        # Convert to the matrix format.
        conmat = {}
        for i in range(0, len(frag)):
            at1 = "ATOM:"+str(i)
            for j in range(0, len(frag)):
                at2 = "ATOM:"+str(j)
                if i == j:
                    continue
                v = round(0.5*(bo[at1][at2] + bo[at2][at1]))
                if v > 0:
                    if i not in conmat:
                        conmat[i] = {}
                    conmat[i][j] = v

        frag.conmat = conmat

    def set_system_connectivity(self, sys, log, kxs=None):
        """
        Compute the connectivty of a given fragment using the fragment bond
        order, and sets that information to frag.conmat.

        Args:
          sys (BigDFT.Fragments.System): the system to compute the
            connectivity of.
          log (Logfile): the Logfile of a calculation of the full system.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
        """
        from scipy.spatial import KDTree
        from BigDFT.Fragments import System, Fragment, GetFragTuple

        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        sys.conmat = {}

        # Create a KD tree of atoms, and a lookup table back to their
        # fragments.
        poslist = []
        frag_lookup = []
        atom_lookup = []
        within_index_lookup = []
        for fragid, frag in sys.items():
            for i, at in enumerate(frag):
                poslist.append(at.get_position())
                frag_lookup.append(fragid)
                atom_lookup.append(at)
                within_index_lookup.append(i)
        tree = KDTree(poslist)

        # Loop over every fragment and atom.
        for fragid, frag in sys.items():
            sys.conmat[fragid] = []
            for at in frag:
                # First, we create a system which is an atom and its
                # nearest neighbors.
                # For these parameters, it's probably fair to say no
                # atom bonds to more than 5 others or to an atom more
                # than 6 bohr away.
                ndist, nearest = tree.query(at.get_position(), k=5+1,
                                            distance_upper_bound=6.0)
                nsys = System()
                nsys["TARGET:0"] = Fragment([at])
                for idx, dist in zip(nearest[1:], ndist[1:]):
                    if dist == float("inf"):
                        break
                    nsys["ATOM:"+str(idx)] = Fragment([atom_lookup[idx]])

                # Compute the bond order which is now an atomic bond order
                # between this atom and its neighbors.
                bo = self.fragment_bond_order(nsys, ["TARGET:0"],
                                              nsys.keys(), log, kxs=kxs)
                bo = bo["TARGET:0"]
                del bo["TARGET:0"]

                toadd = {}
                for k, v in bo.items():
                    rounded = round(v)
                    if rounded > 0:
                        kt = GetFragTuple(k)
                        atom_number = int(kt[1])
                        lookup_key = (frag_lookup[atom_number],
                                      within_index_lookup[atom_number])
                        toadd[lookup_key] = rounded
                sys.conmat[fragid].append(toadd)

    def create_layered_qmmm_sys(self, system, log, target, cutoff, layers,
                                criteria="bondorder", link_atoms=False,
                                kxs=None):
        """
        Creates a multilayered system suitable for QM/MM calculations.
        For each layer, a suitable QM region is built around it.

        Args:
          system (System): a System class, already broken up into fragments.
          log (Logfile): the Logfile of a calculation of the full system.
          target (str): the name of the fragment to treat as the target of
            the qm/mm run.
          cutoff (float): a cutoff value for fragment interactions.
          layers (int): the number of layeres
          criteria (str): how to determine which atoms are included in the
            QM region. Valid choices are "bondorder" and "distance".
          link_atoms (bool): whether to generate link atoms.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.

        Returns:
          (list): a list of Systems, one for each QM layer.
          (System): the MM region.

        """
        from copy import deepcopy
        from BigDFT.Fragments import System

        # First, we will build the layers using only keys.
        qm_list = []
        qm, mm = self.create_qmmm_system(system, log, target, cutoff,
                                         criteria, link_atoms=False, kxs=kxs)
        qm_list.append(list(qm.keys()))
        mm_keys = list(mm.keys())

        # To avoid duplicates
        used_list = deepcopy(list(qm.keys()))

        for l in range(1, layers):
            qm_list.append([])
            for subtarget in qm_list[l-1]:
                qm, mm = self.create_qmmm_system(system, log, subtarget,
                                                 cutoff, criteria,
                                                 link_atoms=False, kxs=kxs)
                qm_list[l] += qm.keys()

            # Remove duplicates
            qm_list[l] = [x for x in qm_list[l] if x not in used_list]
            qm_list[l] = list(set(qm_list[l]))
            used_list = list(set(used_list + qm_list[l]))

            mm_keys = [x for x in mm_keys if x not in qm_list[l]]

        # Next, we will convert those keys into systems.
        qmsys = []
        for l in range(0, layers):
            qmsys.append(System())
            for k in qm_list[l]:
                qmsys[l][k] = deepcopy(system[k])
        mmsys = System()
        for k in mm_keys:
            mmsys[k] = deepcopy(system[k])

        # Add the link atoms.
        if link_atoms:
            # One big system to add the link atoms.
            mergesys = System()
            for l in range(0, layers):
                for fragid, frag in qmsys[l].items():
                    mergesys[fragid] = frag
            qmsys = self.generate_link_atoms(system, mergesys, log, kxs)

            # Break the system up again by layers.
            for l in range(0, layers):
                for fragid in qmsys[l]:
                    qmsys[l][fragid] = mergesys[fragid]

        return qmsys, mmsys

    def create_qmmm_system(self, system, log, target, cutoff,
                           criteria="bondorder", link_atoms=False,
                           kxs=None):
        """
        Creates a system suitable for QM/MM calculations.

        Args:
          system (System): a System class, already broken up into fragments.
          log (Logfile): the Logfile of a calculation of the full system.
          target (str): the name of the fragment to treat as the target of
            the qm/mm run.
          cutoff (float): a cutoff value for fragment interactions.
          criteria (str): how to determine which atoms are included in the
            QM region. Valid choices are "bondorder" and "distance".
          link_atoms (bool): whether to generate link atoms.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.

        Returns:
          (System): the QM region.
          (System): the MM region.
        """
        from BigDFT.Fragments import System, pairwise_distance
        from copy import deepcopy

        qmsys = System()
        mmsys = deepcopy(system)

        if criteria == "bondorder":
            # Compute the spillage
            interaction = self.fragment_bond_order(
                system, [target], system.keys(), log, kxs=kxs)
            interaction = interaction[target]

            # Order the fragments by their spillage value
            sort_spill = sorted(interaction.keys(),
                                key=lambda x: interaction[x], reverse=True)

            # Iterate until we reach the cutoff
            cumsum = sum(interaction.values())
            for key in sort_spill:
                qmsys[key] = deepcopy(system[key])
                del mmsys[key]

                # check cutoff
                cumsum -= interaction[key]
                if cumsum < cutoff:
                    break

        elif criteria == "distance":
            interaction = {x: pairwise_distance(system[target], system[x])
                           for x in system.keys()}
            for fragid, frag in system.items():
                if interaction[fragid] < cutoff:
                    qmsys[fragid] = deepcopy(frag)
                    del mmsys[fragid]

        else:
            raise ValueError("Criteria must be either distance or bondorder")

        if link_atoms:
            if not kxs:
                kxs = self.get_matrix_kxs(log)
            qmsys = self.generate_link_atoms(system, qmsys, log, kxs)

        return qmsys, mmsys

    def fragment_bond_order(self, sys, fraglist1, fraglist2, log, kxs=None,
                            frag_indices=None):
        """
        Computes "bond order" between two sets of fragments using the method of
        Mayer. For two atomic fragments, this would describe the bond
        multiplicity of the covalent bond.

        Args:
          sys (BigDFT.Fragments.System): the system containing the fragments
            of interest
          fraglist1 (list): a list of fragments to compute the bond order
            of.
          fraglist2 (list): a list of fragments to compute the bond order
            between.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
          frag_indices (dict): the matrix indices associated with each
            fragment.

        Return:
          (dict): a dictionary of dictionaries, mapping the bond order of each
          fragment in fraglist1 to each fragment in fraglist2.
        """
        # Get the metadata
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        if frag_indices is None:
            frag_indices = self.get_frag_indices(sys, log)

        # Double loop over the pairs to compute.
        bond_orders = {}
        for fragid1 in fraglist1:
            bond_orders[fragid1] = {}
            idx1 = frag_indices[fragid1]
            smat1 = kxs[:, idx1]

            for fragid2 in fraglist2:
                idx2 = frag_indices[fragid2]
                smat2 = smat1[idx2, :]
                bond_orders[fragid1][fragid2] = sum(x**2 for x in smat2.data)

        return bond_orders

    def fragment_population(self, sys, fragid, log, kxs=None):
        """
        Performs Mulliken population analysis on a fragment, in case charges
        haven't been computed by doing a multipole analysis.

        Args:
          sys (BigDFT.Fragments.System): the system containing the fragment
            of interest
          frag (str): the fragment to compute the charge of.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.

        Return:
          (float): the amount of charge on a fragment.
        """
        from numpy import trace

        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        # Get the metadata
        frag_indices = self.get_frag_indices(sys, log)
        indices = frag_indices[fragid]

        smat = kxs[:, indices]
        smat = smat[indices, :]

        return sum(x.nel for x in sys[fragid]) - trace(smat.todense())

    def generate_link_atoms(self, fullsys, subsys, log, kxs, distcut=6.0):
        """
        This routine adds link atoms to a subsystem based on the bond
        order of a full system.

        Args:
          fullsys (BigDFT.Fragments.System): the full system that the subsystem
            is embedded into.
          subsys (BigDFT.Fragments.System): the embedded system which needs
            link atoms.
          log (BigDFT.Logfiles.Logfile): the log describing a calculation.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
          distcut (float): this cutoff is the largest distance value we expect
            allow a bond to be.

        Returns:
          (BigDFT.Fragments.System): the subsystem with link atoms added.
        """
        from copy import deepcopy

        if fullsys.conmat is None:
            raise ValueError("Generating link atoms requires connectivity"
                             " information")

        linksys = deepcopy(subsys)

        # Loop over atoms and look for bonds running out of the QM system.
        for fragid in subsys:
            linklist = []
            for i in range(0, len(fullsys[fragid])):
                for ft, bv in fullsys.conmat[fragid][i].items():
                    if ft[0] in subsys.keys():
                        continue
                    if bv == 1:
                        newat = deepcopy(fullsys[ft[0]][ft[1]])
                        # Change the symbol to hydrogen
                        newat.sym = "H"
                        linklist.append(newat)
                    elif bv > 1:
                        print("Not yet implemented double/triple bonds.", bv)
                        raise NotImplementedError

            # Add those atoms to the fragment
            for link in linklist:
                linksys[fragid] += [link]

        return linksys

    def run_compute_purity(self, system, log, kxs=None, frag_list=None,
                           frag_indices=None):
        """
        Compute the purity values of the different system fragments.

        Note that this can also be computed using the fragment multipoles,
        but this provides an implementation for when you don't need those
        values.

        Args:
          system (System): instance of a System class, which defines the
            fragmentation.
          log (Logfile): logfile from the run computed on this system.
          kxs (scipy.sparse.csc_matrix): the matrix K*S, which might be already
            computed to reduce I/O time.
          frag_list (list): we can also only compute the purity values of some
            select fragments.
          frag_indices (list): the indices of the matrix associated with each
            fragment. This can be precomputed and passed.

        Returns:
          (dict): for each fragment id, what is the purity value.
        """
        # Handle the optional parameters
        if frag_list is None:
            working_sys = system
        else:
            working_sys = {k: system[k] for k in frag_list}
        if kxs is None:
            kxs = self.get_matrix_kxs(log)

        # Get the metadata
        if frag_indices is None:
            fidx = self.get_frag_indices(working_sys, log)
        else:
            fidx = {x: frag_indices[x] for x in working_sys}

        # Compute the purity array
        charges = {}
        for fragid, frag in working_sys.items():
            charges[fragid] = sum(x.nel for x in frag)
        purity_array = _compute_purity_values(kxs, charges, fidx)

        # Set the purity values of the system
        for fragid in working_sys:
            system[fragid].purity_indicator = purity_array[fragid]

        return purity_array

    def run_compute_spillage(self, system, log, targetid):
        """
        Compute a measure of the spillage interaction between fragments.

        Args:
          system (System): instance of a System class, which defines the
            fragments we will use.
          log (Logfile): instance of a Logfile class
          targetid (str): which fragment to compute the spillage of all other
            fragments with. You can either set this or target.
        Returns:
          (tuple): target spillage value and for each fragment id, what the
          spillage value is.
        """
        from os.path import join, isfile
        from os import environ
        from BigDFT.Spillage import compute_spillage_values
        from BigDFT.Spillage import serial_compute_spillbase
        from scipy.sparse import csc_matrix
        from scipy.io import mmread

        # Define the input files.
        data_dir = _get_datadir(log)
        sfile = join(data_dir, "overlap_sparse.txt")
        hfile = join(data_dir, "hamiltonian_sparse.txt")

        # Convert to text format if necessary
        if log.log["lin_general"]["output_mat"] == 4:
            for fname in ["overlap_sparse.txt", "hamiltonian_sparse.txt"]:
                infile = join(data_dir, fname.replace(".txt", ".mpi"))
                outfile = join(data_dir, fname)
                if not isfile(outfile):
                    self.convert_matrix_format(conversion="binary_to_bigdft",
                                               infile=infile, outfile=outfile)
        # Get the metadata
        frag_indices = self.get_frag_indices(system, log)

        # Check whether to use python or bigpoly version
        # And then perform the computation of the inverse etc
        if isfile(join(environ['BIGDFT_ROOT'], "BigPoly")):
            options = {}
            options["action"] = "compute_spillage"
            options["infile"] = hfile
            options["infile2"] = sfile
            options["outfile"] = join(data_dir, "sinvxhfile.mtx")
            options["outfile2"] = join(data_dir, "sinvxh2file.mtx")

            self.compute_spillage(**options)

            # Read from file
            sinvxh = csc_matrix(mmread(options["outfile"]))
            sinvxh2 = csc_matrix(mmread(options["outfile2"]))
        else:
            # First convert to ccs matrix format
            soutfile = join(data_dir, "overlap_sparse.ccs")
            if not isfile(soutfile):
                self.convert_matrix_format(conversion="bigdft_to_ccs",
                                           infile=sfile, outfile=soutfile)
            houtfile = join(data_dir, "hamiltonian_sparse.ccs")
            if not isfile(houtfile):
                self.convert_matrix_format(conversion="bigdft_to_ccs",
                                           infile=hfile, outfile=houtfile)
            # Then compute with python
            sinvxh, sinvxh2 = serial_compute_spillbase(soutfile, houtfile)

        # Compute the spillage array
        spillage_tuple = compute_spillage_values(
            sinvxh, sinvxh2, frag_indices, targetid)

        return spillage_tuple

    def set_fragment_multipoles(self, system, log):
        """
        Set the fragment multipoles of a system based on a run.

        The multipoles and purity values of the system are
        updated by this call.

        Args:
          system (System): instance of a System class, which defines the
            fragments we will use.
          log (Logfile): instance of a Logfile class
        """
        mp_data = self._run_fragment_multipoles(log, system)

        # Update multipoles and purity values.
        mp_dict = mp_data["Orbital occupation"][0]["Fragment multipoles"]

        for frag, fdata in zip(system.values(), mp_dict):
            frag.purity_indicator = float(fdata["Purity indicator"])
            frag.q0 = [float(x) for x in fdata["q0"]]
            frag.q1 = [float(x) for x in fdata["q1"]]
            frag.q2 = [float(x) for x in fdata["q2"]]

    def get_frag_indices(self, sys, log):
        """
        Compute a lookup table of matrix indices for each fragment in a
        system.

        Args:
          system (System): instance of a System class, which defines the
            fragmentation.
          log (Logfile): logfile from the run computed on this system.

        Returns:
          (dict): a mapping of fragment ids to lists of indices
        """
        from BigDFT.Spillage import MatrixMetadata
        from os.path import join

        data_dir = _get_datadir(log)
        metadatafile = join(data_dir, "sparsematrix_metadata.dat")
        metadata = MatrixMetadata(metadatafile)
        frag_indices = metadata.get_frag_indices(sys)

        return frag_indices

    def get_matrix_kxs(self, log):
        """
        Computes the matrix K*S, the mulliken version of the density matrix,
        and loads it into memory.

        log (Logfile): instance of a Logfile class

        Returns:
          (scipy.sparse.csc_matrix): the matrix K*S
        """
        from os.path import join, isfile
        from os import environ
        from BigDFT.Spillage import serial_compute_puritybase
        from scipy.sparse import csc_matrix
        from scipy.io import mmread
        # Define the input files.
        data_dir = _get_datadir(log)
        sfile = join(data_dir, "overlap_sparse.txt")
        dfile = join(data_dir, "density_kernel_sparse.txt")

        # Convert to text format if necessary
        if log.log["lin_general"]["output_mat"] == 4:
            for fname in ["overlap_sparse.txt", "density_kernel_sparse.txt"]:
                infile = join(data_dir, fname.replace(".txt", ".mpi"))
                outfile = join(data_dir, fname)
                if not isfile(outfile):
                    self.convert_matrix_format(conversion="binary_to_bigdft",
                                               infile=infile, outfile=outfile)

        # Check whether to use python or bigpoly version
        # And then perform the computation of the inverse etc
        if isfile(join(environ['BIGDFT_ROOT'], "BigPoly")):
            options = {}
            options["action"] = "multiply_matrices"
            options["infile"] = sfile
            options["infile2"] = dfile
            options["outfile"] = join(data_dir, "kxs.mtx")

            if not isfile(options["outfile"]):
                self.compute_spillage(**options)

            # Read from file
            kxs = csc_matrix(mmread(options["outfile"]))
        else:  # Compute serially with python
            # First convert to ccs matrix format
            soutfile = join(data_dir, "overlap_sparse.ccs")
            if not isfile(soutfile):
                self.convert_matrix_format(conversion="bigdft_to_ccs",
                                           infile=sfile, outfile=soutfile)
            doutfile = join(data_dir, "density_kernel_sparse.ccs")
            if not isfile(doutfile):
                self.convert_matrix_format(conversion="bigdft_to_ccs",
                                           infile=dfile, outfile=doutfile)
            # Then compute with python
            kxs = serial_compute_puritybase(soutfile, doutfile)

        return kxs

    def _invoke_command(self, command, **kwargs):
        from futile.Utils import option_line_generator
        _system_command(command, option_line_generator(**kwargs))

    def _run_fragment_multipoles(self, log, system=None, orbitals=None):
        """
        Performs the actual run of the fragment multipoles. This means we
        process the default parameters, override the parameters if they are
        specified, write all the necessary input files, and then run.

        Returns:
          (str): the name of the file containing the multipoles
        """
        from os.path import join, isfile
        from os import remove
        from inspect import getargspec
        from yaml import dump
        from futile.YamlIO import load

        # Convert the arguments of the function to a dictionary
        args, vargs, keywords, default = getargspec(self.fragment_multipoles)
        options = {a: d for a, d in zip(args, default)}

        # Use the logfile to determine the right matrix format
        format = log.log["lin_general"]["output_mat"]
        if format == 4:
            options["matrix_format"] = "parallel_mpi-native"
            for key in options:
                if ".txt" in options[key]:
                    options[key] = options[key].replace(".txt", ".mpi")

        # Replace the default directory with the appropriate one if it is
        # available
        if log.log["radical"]:
            data_dir = _get_datadir(log)
            for a, d in options.items():
                if a == "mpirun" or a == "action" or a == "matrix_format":
                    continue
                options[a] = join(data_dir, d)

        # Create the frag.yaml file from the provided system.
        if system:
            system.write_fragfile(options["fragment_file"], log)

        # Create the orbitals.yaml file from the provided orbitals.
        if orbitals is None:
            with open(options["orbital_file"], "w") as ofile:
                ofile.write(dump([-1]))

        if isfile(options["log_file"]):
            remove(options["log_file"])
        self.fragment_multipoles(**options)

        return load(options["log_file"], doc_lists=False)


def _compute_purity_values(kxs, charges, frag_indices):
    """
    Computes the actual purity values.

    Args:
      kxs (scipy.sparse.csc): K*S
      charges (dict): the charge to divide by for each framgent.
      frag_indices (dict): indices associated with each fragment.

    Returns:
      (dict): purity of each fragment.
    """
    from numpy import trace
    purity_values = {}

    for id in frag_indices:
        indices = frag_indices[id]
        smat = kxs[:, indices]
        smat = 0.5 * smat[indices, :]
        purity_values[id] = 2.0 * trace(
            (smat.dot(smat) - smat).todense()) / charges[id]

    return purity_values




import numpy as np
from futile.Utils import write
HaeV = 27.21138386


def _occ_and_virt(log):
    """
    Extract the number of occupied and empty orbitals from a logfile
    """
    norb = log.log['Total Number of Orbitals']
    if log.log['Spin treatment'] == 'Averaged':
        norbv = log.evals[0].info[0]-norb
        return (norb,), (norbv,)
    elif log.log['Spin treatment'] == 'Collinear':
        mpol = log.log['dft']['mpol']
        norbu = int((norb+mpol)/2)
        norbd = norb-norbu
        norbvu = log.evals[0].info[0]-norbu
        norbvd = log.evals[0].info[0]-norbd
        return (norbu, norbd), (norbvu, norbvd)
    else:
        raise ValueError('Information for the orbitals to be implemented')


def transition_indexes(np, nalpha, indexes):
    """
    Returns the list of the indices in the bigdft convention that correspond
    to the couple iorb-ialpha with given spin.

    Args:
        np (tuple): (norbu,norbd) occupied orbitals: when of length 1 assumed
          spin averaged
        nalpha (tuple): (norbu, norbd)virtual orbitals: when of length 1
          assumed spin averaged
        indexes (list): list of tuples of (iorb,ialpha,ispin) desired indices
          in python convention (start from 0)
    """
    nspin = len(np)
    inds = []
    for iorb, ialpha, ispin in indexes:
        jspin = ispin if nspin == 2 else 0
        ind = ialpha+iorb*nalpha[jspin]  # local index of the spin subspace
        if ispin == 1:
            ind += np[0]*nalpha[0]  # spin 2 comes after spin one
        inds.append(ind)
    return inds


def _collection_indexes(np, nvirt_small):
    harvest = []
    for ispin in [0, 1]:
        jspin = ispin if len(np) == 2 else 0
        for ip in range(np[jspin]):
            for ialpha in range(nvirt_small[jspin]):
                harvest.append([ip, ialpha, ispin])
    return harvest


def _collection_indexes_iocc(iocc, nvirt, spin=None):
    """
    For each iocc and a selected spin provide the list of couples that are
    concerned up to nvirt
    If spin is none provide the list for all values of the spin
    """
    harvest = []
    for ispin in [0, 1]:
        jspin = ispin if len(nvirt) == 2 else 0
        if spin is not None and ispin != spin:
            continue
        for ialpha in range(nvirt[jspin]):
            harvest.append([iocc, ialpha, ispin])
    return harvest


class TransitionMatrix(np.matrix):
    """
    Matrix of Transition Quantities, might be either :class:`CouplingMatrix`
    or :class:`TransitionMultipoles`

    Args:
        matrix (matrix-like): data of the coupling matrix. If present also
          the number of orbitals should be provided.
        norb_occ (tuple): number of occupied orbitals per spin channnel.
          Compulsory if ``matrix`` is specified.
        norb_virt (tuple): number of empty orbitals per spin channnel.
          Compulsory if ``matrix`` is specified.
        log (Logfile): Instance of the logfile from which the coupling matrix
          calculation is performed. Automatically retrieves the ``norb_occ``
          and `norb_virt`` parameters. When ``log`` parameter is present the
          parameter ``matrix`` is ignored.

    Raises:
        ValueError: if the file of the coupling matrix indicated by ``log``
          does not exists
    """
    def __new__(cls, matrix=None, norb_occ=None, norb_virt=None, log=None):
        """
        Create the object from the arguments and return the ``self`` instance
        """
        import os
        if log is not None:
            datadir = log.log.get('radical', '')
            datadir = 'data-'+datadir if len(datadir) > 0 else 'data'
            cmfile = os.path.join(log.srcdir, datadir, cls._filename)
            if not os.path.isfile(cmfile):
                raise ValueError('The file "'+cmfile+'" does not exist')
            norb, norbv = _occ_and_virt(log)
            write('Loading data with ', norb, ' occupied and ',
                  norbv, ' empty states, from file "', cmfile, '"')
            try:
                import pandas as pd
                write('Using pandas:')
                mat = pd.read_csv(cmfile, delim_whitespace=True, header=None)
            except ImportError:
                write('Using numpy:')
                mat = np.loadtxt(cmfile)
            write('done')
        else:
            mat = matrix
        return super(TransitionMatrix, cls).__new__(cls, mat)

    def __init__(self, *args, **kwargs):
        """
        Perform sanity checks on the loaded matrix
        """
        log = kwargs.get('log')
        if log is not None:
            self.norb_occ, self.norb_virt = _occ_and_virt(log)
        else:
            self.norb_occ = kwargs.get('norb_occ')
            self.norb_virt = kwargs.get('norb_virt')
        assert(self.shape[0] == self._total_transitions())
        write("Shape is conformal with the number of orbitals")
        self._sanity_check()

    def _total_transitions(self):
        ntot = 0
        for no, nv in zip(self.norb_occ, self.norb_virt):
            ntot += no*nv
        if len(self.norb_occ) == 1:
            ntot *= 2
        return ntot

    def _subindices(self, norb_occ, norb_virt):
        for i, (no, nv) in enumerate(zip(norb_occ, norb_virt)):
            assert(no <= self.norb_occ[i] and nv <= self.norb_virt[i])
        harvest = _collection_indexes(norb_occ, norb_virt)
        return np.array(transition_indexes(norb_occ, self.norb_virt, harvest))

    def _sanity_check(self):
        pass


class CouplingMatrix(TransitionMatrix):
    """
    Casida Coupling Matrix, extracted from the calculation performed by BigDFT
    """
    _filename = 'coupling_matrix.txt'

    def _sanity_check(self):
        write('Casida Matrix is symmetric',
              np.allclose(self, self.T, atol=1.e-12))

    def subportion(self, norb_occ, norb_virt):
        """Extract a subportion of the coupling matrix.

        Returns a Coupling Matrix which is made by only considering the first
        ``norb_occ`` and ``norb_virt`` orbitals

        Args:
           norb_occ (tuple): new number of occupied orbitals. Must be lower
             that the instance value
           norb_virt (tuple): new number of virtual orbitals. Must be lower
             that the instance value
        """
        inds = self._subindices(norb_occ, norb_virt)
        mat = np.array([row[0, inds] for row in self[inds]])
        return CouplingMatrix(matrix=mat, norb_occ=norb_occ,
                              norb_virt=norb_virt)

    def diagonalize(self):
        """
        Diagonalize the Coupling Matrix

        Returns:
            (np.matrix, np.matrix):
            tuple of the Eigenvvalues and Eigenvectors of the coupling matrix,
              as returned by :meth:`numpy.linalg.eigh`. We perform the
              transpose of the matrix with eigenvectors to have them sorted as
              row vectors
        """
        write('Diagonalizing Coupling matrix of shape', self.shape)
        E2, C_E2 = np.linalg.eigh(self)
        write('Eigensystem solved')
        C_E2 = C_E2.T
        return E2, C_E2


class TransitionMultipoles(TransitionMatrix):
    """
    Transition dipoles, extracted from the calculation performed by BigDFT
    """
    _filename = 'transition_quantities.txt'

    def subportion(self, norb_occ, norb_virt):
        """Extract a subportion of the Transition Multipoles.

        Returns a set of transition multipoles which is made by only
        considering the first ``norb_occ`` and ``norb_virt`` orbitals

        Args:
           norb_occ (tuple): new number of occupied orbitals. Must be lower
             that the instance value
           norb_virt (tuple): new number of virtual orbitals. Must be lower
             that the instance value

        Returns:
           TransitionMultipoles: reduced transition multipoles
        """
        inds = self._subindices(norb_occ, norb_virt)
        mat = np.array(self[inds])
        return TransitionMultipoles(matrix=mat, norb_occ=norb_occ,
                                    norb_virt=norb_virt)

    def get_transitions(self):
        """
        Get the transition quantities as the dimensional objects which should
        contribute to the oscillator strengths.

        Returns:
            numpy.array: Transition quantities multiplied by the square root of
            the unperturbed transition energy
        """
        newdipole = []
        for line in self:
            newdipole.append(np.ravel(line[0, 0]*line[0, 1:]))
        return np.array(newdipole)


class TransitionDipoles(TransitionMultipoles):
    """
    Transition dipoles as provided in the version of the code < 1.8.0.
    Deprecated, to be used in some particular cases
    """
    _filename = 'transition_dipoles.txt'

    def get_transitions(self):
        return self


class Excitations():
    """LR Excited states of a system

    Definition of the excited states in the Casida Formalism

    Args:
       cm (CouplingMatrix): the matrix of coupling among transitions
       tm (TransitionMultipoles): scalar product of multipoles among
         transitions

    """

    def __init__(self, cm, tm):
        self.cm = cm
        self.tm = tm
        self.eigenvalues, self.eigenvectors = cm.diagonalize()

        # : array: transition quantities coming from the multipoles
        self.transitions = tm.get_transitions()

        scpr = np.array(np.dot(self.eigenvectors, self.transitions))

        #: array: oscillator strenghts components of the transitions defined
        #  as the square of  $\int w_a(\mathbf r) r_i $
        self.oscillator_strenghts = np.array([t**2 for t in scpr[:, 0:3]])

        # : array: average of all the components of the OS
        self.avg_os = np.average(self.oscillator_strenghts, axis=1)

        self.alpha_prime = 2.0*self.oscillator_strenghts / \
            self.eigenvalues[:, np.newaxis]
        """ array: elements of the integrand of the statical polarizability in
            the space of the excitations """

        self._indices_for_spin_comparison = \
            self._get_indices_for_spin_comparison()

        self.identify_singlet_and_triplets(1.e-5)

    def _get_indices_for_spin_comparison(self):
        inds = [[], []]
        inds0 = []
        # get the indices for comparison, take the minimum between the spins
        if len(self.cm.norb_occ) == 1:
            nocc = self.cm.norb_occ[0]
            nvirt = self.cm.norb_virt[0]
            nos = [nocc, nocc]
            nvs = [nvirt, nvirt]
        else:
            nocc = min(self.cm.norb_occ)
            nvirt = min(self.cm.norb_virt)
            nos = self.cm.norb_occ
            nvs = self.cm.norb_virt
        for ispin in [0, 1]:
            for a in range(nvirt):
                for p in range(nocc):
                    inds[ispin].append([p, a, ispin])
            for a in range(nvirt, nvs[ispin]):
                for p in range(nocc, nos[ispin]):
                    inds0.append([p, a, ispin])
        transA = transition_indexes(
            self.cm.norb_occ, self.cm.norb_virt, inds[0])
        transB = transition_indexes(
            self.cm.norb_occ, self.cm.norb_virt, inds[1])
        trans0 = transition_indexes(self.cm.norb_occ, self.cm.norb_virt, inds0)
        return transA, transB, trans0

    def spectrum_curves(self, omega, slice=None, weights=None):
        """Calculate spectrum curves.

        Provide the set of the curves associated to the weights. The resulting
        curves might then be used to draw the excitation spectra.

        Args:
            omega (array): the linspace used for the plotting, of shape
              ``(n,)``. Must be provided in Atomic Units
            slice (array): the lookup array that has to be considered. if Not
              provided the entire range is assumed
            weights (array): the set of arrays used to weight the spectra. Must
              have shape ``(rank,m)``, where ``rank`` is equal to the number
              of eigenvalues. If m is absent it is assumed to be 1. When not
              specified, it defaults to the average oscillator strenghts.

        Returns:
            array: a set of spectrum curves, of shape equal to ``(n,m)``,
            where ``n`` is the shape of ``omega`` and ``m`` the size of the
            second dimension of ``weights``.
        """
        if slice is None:
            oo = self.eigenvalues[:, np.newaxis] - omega**2
            wgts = weights if weights is not None else self.avg_os
        else:
            oo = self.eigenvalues[slice, np.newaxis] - omega**2
            oo = oo[0]
            wgts = weights if weights is not None else self.avg_os[slice]
        return np.dot(2.0/oo.T, wgts)

    def identify_singlet_and_triplets(self, tol=1.e-5):
        """
        Find the lookup tables that select the singlets and the triplets among
          the excitations

        Args:
           tol (float): tolerance to be applied to recognise the spin character
        """
        sings = []
        trips = []
        for exc in range(len(self.eigenvalues)):
            sing, trip = self.project_on_spin(exc, tol)
            if sing:
                sings.append(exc)
            if trip:
                trips.append(exc)

        if len(sings) > 0:
            self.singlets = (np.array(sings),)
            """array:  lookup table of the singlet excitations"""

        if len(trips) > 0:
            self.triplets = (np.array(trips),)
            """array:  lookup table of the triplet excitations"""

    def _project_on_occ(self, exc):
        """
        Project a given eigenvector on the occupied orbitals.
        In the spin averaged case consider all the spin indices nonetheless
        """
        norb_occ = self.cm.norb_occ
        norb_virt = self.cm.norb_virt
        pProj_spin = []
        for ispin, norb in enumerate(norb_occ):
            pProj = np.zeros(norb)
            for iorb in range(norb):
                harvest = _collection_indexes_iocc(
                    iorb, self.cm.norb_virt, spin=None if len(norb_occ) == 1
                    else ispin)
                inds = np.array(transition_indexes(
                    norb_occ, norb_virt, harvest))
                pProj[iorb] = np.sum(np.ravel(self.eigenvectors[exc, inds])**2)
            pProj_spin.append(pProj)
        return pProj_spin

    def project_on_spin(self, exc, tol=1.e-8):
        """
        Control if an excitation has a Singlet or Triplet character

        Args:
            exc (int): index of the excitation to be controlled

        Returns:
            tuple (bool,bool): couple of booleans indicating if the excitation
            is a singlet or a triplet respectively
        """
        A, B, zero = [np.ravel(self.eigenvectors[exc, ind])
                      for ind in self._indices_for_spin_comparison]
        issinglet = np.linalg.norm(A-B) < tol
        istriplet = np.linalg.norm(A+B) < tol
        return issinglet, istriplet
        # print (self.eigenvalues[exc], np.linalg.norm(A), np.linalg.norm(B),
        #        A-B,A+B, np.linalg.norm(zero))

    def _get_threshold(self, pProj_spin, th_energies, tol):
        """
        Identify the energy which is associated to the threshold of a given
        excitation. The tolerance is used to discriminate the component
        """
        ths = -1.e100
        for proj, en in zip(pProj_spin, th_energies):
            norb = len(en)
            pProj = proj.tolist()
            pProj.reverse()
            imax = norb-1
            for val in pProj:
                if val > tol:
                    break
                imax -= 1
            ths = max(ths, en[imax])
        return ths

    def split_excitations(self, evals, tol, nexc='all'):
        """Separate the excitations in channels.

        This methods classify the excitations according to the channel they
        belong, and determine if a given excitation might be considered as a
        belonging to a discrete part of the spectrum or not.

        Args:
            evals (BandArray): the eigenvalues as they are provided
              (for instance) from a `Logfile` class instance.
            tol (float): tolerance for determining the threshold
            nexc (int,str): number of excitations to be analyzed.
                If ``'all'`` then the entire set of excitations are analyzed.

        """
        self.determine_occ_energies(evals)
        self.identify_thresholds(self.occ_energies, tol, len(
            self.eigenvalues) if nexc == 'all' else nexc)

    def identify_thresholds(self, occ_energies, tol, nexc):
        """Identify the thresholds per excitation.

        For each of the first ``nexc`` excitations, identify the energy value
        of its corresponding threshold. This value is determined by projecting
        the excitation components on the occupied states and verifying that
        their norm for the highest energy level is below a given tolerance.

        Args:
           occ_energies (tuple of array-like): contains the list of the
             energies of the occupied states per spin channel
           tol (float): tolerance for determining the threshold
           nexc (int): number of excitations to be analyzed
        """
        # : Norm of the $w_p^a$ states associated to each excitation
        self.wp_norms = []
        threshold_energies = []
        for exc in range(nexc):
            proj = self._project_on_occ(exc)
            self.wp_norms.append(proj)
            threshold_energies.append(
                self._get_threshold(proj, occ_energies, tol))
        # : list: identified threshold for inspected excitations
        self.threshold_energies = np.array(threshold_energies)

        self.excitations_below_threshold = np.where(
            np.abs(self.threshold_energies) >= np.sqrt(
                self.eigenvalues[0:len(self.threshold_energies)]))
        """ array: Indices of the excitations which lie below their
                    corresponding threshold """

        self.excitations_above_threshold = np.where(
            np.abs(self.threshold_energies) <
            np.sqrt(self.eigenvalues[0:len(self.threshold_energies)]))
        """ array: Indices of the excitations which lie above their
                   corresponding threshold """

    def determine_occ_energies(self, evals):
        """
        Extract the occupied energy levels from a Logfile BandArray structure,
        provided the tuple of the number of occupied states

        Args:
            evals (BandArray): the eigenvalues as they are provided
              (for instance) from a `Logfile` class instance.
        """
        norb_occ = self.cm.norb_occ
        occ_energies = []
        # istart=0
        for ispin, norb in enumerate(norb_occ):  # range(len(norb_occ)):
            # istart:istart+norb_occ[ispin]]))
            occ_energies.append(np.array(evals[0][ispin][0:norb]))
            # istart+=norb_tot[ispin]
        # : array: energies of the occupied states out of the logfile
        self.occ_energies = occ_energies
        # : float: lowest threshold of the excitations. All excitations are
        #   discrete below this level
        self.first_threshold = abs(
            max(np.max(self.occ_energies[0]), np.max(self.occ_energies[-1])))

    def plot_alpha(self, **kwargs):
        """Plot the imaginary part.

        Plot the real or imaginary part of the dynamical polarizability.

        Keyword Arguments:
           real (bool): True if real part has to be plotted. The imaginary
             part is plotted otherwise
           eta (float): Value of the complex imaginary part. Defaults to 1.e-2.
           group (str): see :meth:`lookup`

           **kwargs:
               other arguments that might be passed to the :meth:`plot` method
               of the :mod:`matplotlib.pyplot` module.

        Returns:
             :mod:`matplotlib.pyplot`: the reference to
             :mod:`matplotlib.pyplot` module.
        """
        import matplotlib.pyplot as plt
        from futile.Utils import kw_pop
        emax = np.max(np.sqrt(self.eigenvalues))*HaeV
        kwargs, real = kw_pop('real', False, **kwargs)
        plt.xlim(xmax=emax)
        if real:
            plt.ylabel(r'$\mathrm{Re} \alpha$ (AU)', size=14)
        else:
            plt.ylabel(r'$\mathrm{Im} \alpha$', size=14)
            plt.yticks([])
        plt.xlabel(r'$\omega$ (eV)', size=14)
        if hasattr(self, 'first_threshold'):
            eps_h = self.first_threshold*HaeV
            plt.axvline(x=eps_h, color='black', linestyle='--')
        kwargs, eta = kw_pop('eta', 1.e-2, **kwargs)
        omega = np.linspace(0.0, emax, 5000)+2.0*eta*1j
        kwargs, group = kw_pop('group', 'all', **kwargs)
        slice = self.lookup(group)
        spectrum = self.spectrum_curves(omega, slice=slice)
        toplt = spectrum.real if real else spectrum.imag
        pltkwargs = dict(c='black', linewidth=1.5)
        pltkwargs.update(kwargs)
        plt.plot(omega*HaeV, toplt, **pltkwargs)
        return plt

    def lookup(self, group):
        """
        Identify the group of the excitations according to the argument

        Args:
            group (str):
               A string chosen between

               * ``"all"`` : provides the entire set of excitations
                 (:py:class:`None` instead of the lookup array)

               * ``"bt"`` : provides only the excitations below threshold

               * ``"at"`` : provides only the excitations above threshold

               * ``"singlets"`` : provides the index of the excitations that
                 have a singlet character

               * ``"triplets"`` : provides the index of the excitations that
                 have a triplet character

        """
        slice = None
        if group == 'bt':
            slice = self.excitations_below_threshold
        if group == 'at':
            slice = self.excitations_above_threshold
        if group == 'singlets':
            slice = self.singlets
        if group == 'triplets':
            slice = self.triplets
        return slice

    def plot_excitation_landscape(self, **kwargs):
        """
        Represent the excitation landscape as splitted in the excitations class

        Args:
            **kwargs:
               keyword arguments to be passed to the `pyplot` instance.
               The ``xlabel``, ``ylabel`` as well as ``xlim`` are already set.

        Returns:
            :mod:`matplotlib.pyplot`: the reference to :mod:`matplotlib.pyplot`
              module.

        Example:
           >>> ex=Excitations(cm,tm)
           >>> ex.split_excitations(evals=...,tol=1.e-4,nexc=...)
           >>> ex.plot_excitation_landscape(title='Excitation landscape')
        """
        import matplotlib.pyplot as plt
        Emin = 0.0
        Emax = np.max(np.sqrt(self.eigenvalues))*HaeV
        for level in self.occ_energies[0]:
            eng_th = level*HaeV
            plt.plot((Emin, eng_th), (level, level),
                     '--', c='red', linewidth=1)
            plt.plot((eng_th, Emax), (level, level), '-', c='red', linewidth=1)
            plt.scatter(abs(eng_th), level, marker='x', c='red')
        ind_bt = self.excitations_below_threshold
        exc_bt = np.sqrt(self.eigenvalues)[ind_bt]
        lev_bt = self.threshold_energies[ind_bt]
        plt.scatter(HaeV*exc_bt, lev_bt, s=16, marker='o', c='black')
        ind_at = self.excitations_above_threshold
        exc_at = np.sqrt(self.eigenvalues)[ind_at]
        lev_at = self.threshold_energies[ind_at]
        plt.scatter(HaeV*exc_at, lev_at, s=14, marker='s', c='blue')
        plt.xlabel('energy (eV)')
        plt.ylabel('Threshold energy (Ha)')
        plt.xlim(xmin=Emin-1, xmax=Emax)
        for attr, val in kwargs.items():
            if type(val) == dict:
                getattr(plt, attr)(**val)
            else:
                getattr(plt, attr)(val)
        return plt

    def dos_dict(self, group='all'):
        """Dictionary for DoS creation.

        Creates the keyword arguments that have to be passed to the
        `meth:BigDFT.DoS.append` method of the `DoS` class

        Args:
           group (str): see :meth:`lookup`

        Returns:
            :py:class:`dict`: kewyord arguments that can be passed to the
            `meth:BigDFT.DoS.append` method of the :class:`DoS.DoS` class

        """
        ev = np.sqrt(self.eigenvalues)
        slice = self.lookup(group)
        if slice is not None:
            ev = ev[slice]
        return dict(energies=np.array([np.ravel(ev)]), units='AU')

    def dos(self, group='all', **kwargs):
        """Density of States of the Excitations.

        Provides an instance of the :class:`~BigDFT.DoS.DoS` class,
        corresponding to the Excitations instance.

        Args:
           group (str): see :meth:`lookup`

           **kwargs: other arguments that might be passed to the
             :class:`DoS.DoS` instantiation

        Returns:
            :class:`DoS.DoS`: instance of the Density of States class
        """
        from BigDFT.DoS import DoS
        kwa = self.dos_dict(group=group)
        kwa['energies'] = kwa['energies'][0]
        if hasattr(self, 'first_threshold'):
            kwa['fermi_level'] = self.first_threshold
        else:
            kwa['fermi_level'] = 0.0
        kwa.update(kwargs)
        return DoS(**kwa)

    def plot_Sminustwo(self, coord, alpha_ref=None, group='all'):
        """Inspect S-2 sum rule.

        Provides an handle to the plotting of the $S^{-2}$ sum rule, which
        should provide reference values for the static polarizability tensor.

        Args:
            coord (str): the coordinate used for inspection. May be ``'x'``,
              ``'y'`` or ``'z'``.
            alpha_ref (list): diagonal of the reference static polarizability
               tensor (for instance calculated via finite differences).
               If present the repartition of the contribution of the various
               groups of excitations is plotted.
            group (str): see :meth:`lookup`

        Returns:
            reference to :mod:`matplotlib.pyplot` module.

        """
        import matplotlib.pyplot as plt
        idir = ['x', 'y', 'z'].index(coord)
        fig, ax1 = plt.subplots()
        ax1.set_xlabel('energy (eV)', size=14)
        plt.ylabel(r'$\alpha_{'+coord+coord+r'}$ (AU)', size=14)
        if alpha_ref is not None:
            plt.axhline(y=alpha_ref[idir], color='r', linestyle='--')
        if hasattr(self, 'first_threshold'):
            eps_h = abs(HaeV*self.first_threshold)
            plt.axvline(x=eps_h, color='black', linestyle='--')
        e = np.sqrt(self.eigenvalues)*HaeV
        w_ii = self.alpha_prime[:, idir]
        slice = self.lookup(group)
        if slice is not None:
            e = e[slice]
            w_ii = w_ii[slice]
        ax1.plot(e, np.cumsum(w_ii))
        ax2 = ax1.twinx()
        ax2.plot(e, w_ii, color='grey', linestyle='-')
        plt.ylabel(r'$w_{'+coord+coord+r'}$ (AU)', size=14)

        return plt


def get_alpha_energy(log, norb, nalpha):
    return log.evals[0][0][norb+nalpha-1]


def identify_contributions(numOrb, na, exc, C_E2):
    pProj = np.zeros(numOrb*2)
    for p in range(numOrb):
        for spin in [0, 1]:
            # sum over all the virtual orbital and spin
            for alpha in range(na):
                # extract the value of the index of C_E2
                elements = transition_indexes(
                    [numOrb], [na], [[p, alpha, spin]])
                for el in elements:
                    pProj[p+numOrb*spin] += C_E2[exc][el]**2
    pProj = pProj[0:numOrb]+pProj[numOrb:2*numOrb]  # halves the components
    return pProj


def get_p_energy(log, norb):
    return log.evals[0][0][0:norb]


def get_threshold(pProj, th_energies, th_levels, tol):
    norb = len(th_energies)
    pProj = pProj.tolist()
    pProj.reverse()
    imax = norb-1
    for val in pProj:
        if val > tol:
            break
        imax -= 1
    return [th_levels[imax], th_energies[imax]]



#defines the class of the polar plot
class polar_axis():
  step=5
  def __init__(self,fig,ax,data):
    self.fig = fig #pylab.figure()
    self.ax = ax #pylab.axes([0.025,0.025,0.95,0.95], polar=True)
    self.data=data
    self.reset=True
    self.ax.cla()
    datatmp=self.data
    self.subdata=self.dump_timing_level(datatmp)
    self.draw_polarplot()
    self.fig.canvas.mpl_connect('motion_notify_event',self._info_callback)
    #self.fig.canvas.mpl_connect('pick_event',self._new_level)
    self.fig.canvas.mpl_connect('button_press_event',self._new_level)

  def draw_polarplot(self):
    import pylab
    import numpy as np
    #self.subdata=data
    self.tot=self.subdata["time"][0]
    self.times=self.subdata["time"]
    self.N=len(self.times)

    self.width=np.array(self.subdata["time"])/self.tot*2*np.pi
    self.theta=np.array(self.subdata["theta"])/self.tot*2*np.pi
    self.bot=np.array(self.subdata["level"])
    self.names=self.subdata["names"]
    self.radii = np.array(self.N*[self.step])

    self.bars = self.ax.bar(self.theta,self.radii,
                          width=self.width,
                          bottom=self.step*self.bot,picker=True)

    ilev=0
    #maxlev=max(self.bot)
    for r,bar,ilev in zip(self.radii, self.bars,self.theta):
       #print ilev,'hello',float(ilev)/float(N),maxlev
       #bar.set_facecolor( pylab.cm.jet(float(ilev)/maxlev))
       bar.set_facecolor(pylab.cm.jet(float(ilev)/(2.0*np.pi)))
       bar.set_alpha(0.5)
       ilev+=1

    self.ax.set_xticklabels([])
    self.ax.set_yticklabels([])
    self._wrotesome=False
    self.info=self.ax.text(0,0,'',fontsize=150)

  ## to be used what it is the moment to show the plot(s)
  #def show(self):
  #  import pylab
  #  try:
  #    pylab.show()
  #  except KeyboardInterrupt:
  #    raise

  def _find_name(self,th,level):
    import numpy as np
    levth=[]
    ipiv=[]
    for i in range(self.N):
      if self.bot[i]==level:
        levth.append(self.theta[i])
        ipiv.append(i)
    if len(levth)==0: return None
    to_find_zero=np.array(levth)-th
    to_find_zero[to_find_zero > 0]=-4*np.pi
    routine=np.argmin(-to_find_zero)
    if routine == len(to_find_zero)-1 and abs(routine)> 0.5:
      return None
    else:
      return (self.theta[ipiv[routine]],self.names[level][routine],self.times[ipiv[routine]])
    #if min(abs(to_find_zero))==0.0:
    #  routine=np.argmin(abs(to_find_zero))
    #  return (th,self.names[level][routine],self.times[ipiv[routine]])
    #else:
    #  to_find_zero[to_find_zero > 0]=-4*np.pi
    #  routine=np.argmin(-to_find_zero)
    #  return (self.theta[ipiv[routine]],self.names[level][routine],self.times[ipiv[routine]])


  def _info_string(self,xdata,level):
    #create string to be plotted in plot
    (tt,name,time)=self._find_name(xdata,level)
    info=name+':\n'+str(time)+" s ("
    info+= " %6.2f %s" % (time/self.tot*100.0,"% ) \n")
    #find lower level
    it=range(level)
    it.reverse()
    parents=''
    for i in it:
      (tt,name,time)=self._find_name(tt,i)
      parents+="-"+name+"("+str(time)+"s,"
      parents+= " %6.2f %s" % (time/self.tot*100.0,"% ) \n")
    if len(parents) > 0:
      info+="Calling path:\n"+parents
    return info

  def _new_level(self,event):
    #click=event.mouseevent.button
    click=event.button
    datatmp=self.data
    if click==3:
      if not self.reset:
        self.ax.texts.remove(self.comebackinfo)
        self.reset=True
      nameandlev=None
    elif click==1:
      #thisline = event.artist
      #xdata, ydata = thisline.get_xy()
      xdata, ydata=event.xdata,event.ydata
      level = int(ydata)/self.step
      goodpoint=self._find_name(xdata,level)
      if goodpoint is None: return
      (tt,name,time)=goodpoint
      if not self.reset: return #no double level
      self.reset=False
      nameandlev=(name,level)
      #extract the data to be plotted for the internal level
    else:
      return
    self.subdata=self.dump_timing_level(datatmp,starting_point=nameandlev)
    self.ax.cla()
    self.draw_polarplot()
    if not self.reset:
      self.comebackinfo=self.ax.text(0.02,0.95,
                                     'Subroutines of '+name+' (Right click to reset)',
                                     fontsize=15,transform = self.ax.transAxes)
    self.fig.canvas.draw()

  def _info_callback(self,event):
    #print dir(event)
    #thisline = event.artist
    #xdata, ydata = thisline.get_xy()
    xdata=event.xdata#data
    ydata=event.ydata#data
    #print 'test',xdata,ydata,event.x,event.y
    if xdata is None or ydata is None: return
    level = int(ydata)/self.step
    goodpoint=self._find_name(xdata,level)

    #then plot the string
    if self._wrotesome:
      self.ax.texts.remove(self.info)
      self.info=self.ax.text(0,0,'',fontsize=150)
      self._wrotesome=False
    if goodpoint is None: return
    offset = 0.02
    #print 'nowinfo',info
    self.info= self.ax.text( offset, offset, self._info_string(xdata,level),
                             fontsize = 15,transform = self.ax.transAxes )
    self._wrotesome=True
    self.fig.canvas.draw()

  def dump_timing_level(self,level,starting_point=None,ilev=0,theta=0,data=None): ##data={"time":[],"level":[],"theta":[],"names":[[]]},reset=False):
    """Inspect the first level of the given dictionary and dump the profile subroutines at this level"""
    import pylab
    if data is None: data={"time":[],"level":[],"theta":[],"names":[[]]}
    subs=data["time"]
    tht=data["theta"]
    lev=data["level"]
    nms=data["names"]
    if ilev == len(nms):
      nms.append([])
    #tel=0
    tel=theta #entry point of the routine
    t0=0
    start=starting_point
    import copy
    leveltmp=copy.deepcopy(level)
    brk=False
    for routine in leveltmp:
      #first eliminate the subroutines from the level
      try:
        sublevel=routine.pop("Subroutines")
      except:
        sublevel=None
      for name in routine:
        if starting_point is not None:
          if (name,ilev) != starting_point:
            continue
          else:
            brk=True
            start=None #from now on take the data
            ilev=0
        t0=routine[name][0] #take the time
        subs.append(t0) #take the time
        tht.append(tel)
        lev.append(ilev)
        nms[ilev].append(name)
      if sublevel is not None:
        jlev=ilev+1
        self.dump_timing_level(sublevel,starting_point=start,ilev=jlev,theta=tel,data=data)
      tel=tel+t0
      if brk: break
    return data

class TimeData:
  barwidth=0.9#0.35
  ignored_counters=['CPU parallelism','Routines timing and number of calls','SUMMARY','Report timestamp','Hostnames']
  def __init__(self,*filenames,**kwargs):
    """
    Class to analyze the timing from a Futile run
    ------
    Arguments
    filenames: list of the files which have to be treated
    plottype: Decide the default yscale for the plotting
              May be "Percent" or "Seconds"
    static: Show the plot statically for screenshot use
    fontsize: Determine fontsize of the bar chart plot
    nokey: Remove the visualization of the key from the main plot
    counter: retrieve as reference value the counter provided
    only_last: retrieve only the last document in the time.yaml
    """
    #here a try-catch section should be added for multiple documents
    #if (len(filename) > 1
    from futile import Figures as Fig
    only_last = kwargs.get('only_last',False)
    self.log=[]
    for filename in filenames:
      from futile import YamlIO as Yaml
      tmplogs=Yaml.load(filename,doc_lists=True)
      if only_last:
        self.log+=[[ a for a in tmplogs][-1]]
      else:
        self.log+=tmplogs
      #try:
      #  self.log+=[yaml.load(open(filename, "r").read(), Loader = yaml.CLoader)]
      #except:
      #  self.log+=yaml.load_all(open(filename, "r").read(), Loader = yaml.CLoader)
    #create the figure environemnt
    self.figures=Fig.FigureSet(title='Profiling',QuitButton=True,twinaxes=True)
    #self.barfig = None
    #self.axbars = None
    #self.newfigs =[]
    self.radio = None
    self.toggle_unbalancing = False
    self.quitButton = None
    self.lined=[]
    self.plot_start=kwargs.get('plottype','Seconds')
    self.static = kwargs.get('static',False)
    self.fontsize=kwargs.get('fontsize',15)
    self.nokey=kwargs.get('nokey',False)
    self.strong_scaling=kwargs.get('strong_scaling',True)
    counter=kwargs.get('counter','WFN_OPT') #the default value, to be customized
    if counter in self.counters():
      self.inspect_counter(counter)
    else:
      print("Warning: counter not initialized, check the available counters")

  def counters(self):
    "Inspect the available counters"
    cnts=[]
    for doc in self.log:
      if doc is None: continue
      for internal_cnts in doc.keys():
        if internal_cnts not in cnts and internal_cnts not in self.ignored_counters: cnts.append(internal_cnts)
    return cnts

  def _refill_classes(self,main):
    self.classes=[]
    self.totals=[]
    for doc in self.log:
        if doc is None: continue
        scf=doc.get(main)
        if scf is None: continue
        self.scf.append(scf)
        self.routines.append(doc.get("Routines timing and number of calls"))
        self.hostnames.append(doc.get("Hostnames"))
        self.unbalancings.append(True)
        loclass=scf["Classes"].keys()
        if 'Total' in loclass:
          self.totals.append(scf["Classes"]["Total"][1])
        for cs in loclass:
          if cs not in self.classes and cs != "Total": self.classes.append(cs)
          if len(scf["Classes"][cs]) == 2: self.unbalancings[-1]=False
        mpit=doc.get("CPU parallelism")
        title=doc.get("Run name","Unknown")
        if mpit is not None:
            mpi=mpit.get('MPI tasks')
            omp=mpit.get('OMP threads')
            if title is "Unknown":
              title=str(mpi) if omp is None else str(mpi)+'-'+str(omp)
            ncores=mpi
            if omp is not None: ncores*=omp
        else:
            ncores=0.0
        self.ids.append(title)
        self.ncores.append(ncores)
    self.classes.sort()
    self.classes.append("Unknown") #an evergreen

  def inspect_counter(self,counter,unit=None):
    self.routines=[]
    self.hostnames=[]
    self.unbalancings=[]
    self.scf=[]
    self.ids=[]
    self.ncores=[] #number of cores for each of the valid run
    self.vals=self.plot_start if unit is None else unit
    self._refill_classes(counter)
    #here we might add the policy to add new figures or delete previous ones
    data=self.collect_categories(self.scf,self.vals)
    fig,axis=self.figures[0]
    axis.cla()
    self.draw_barfigure(fig,axis,data,title="Counter "+counter)

  def inspect_category(self,cat):
    data=self.find_items(cat,self.scf)
    if self.figures.exists(cat):
      newfig,newax=self.figures[cat]
      #erase the plot in case it is already available
      newax.cla()
    else:
      newfig,newax=self.figures.add(title=cat,twinaxes=True)
    self.draw_barfigure(newfig,newax,data,title=cat)
    newfig.show()

  def show(self):
    self.figures.show()

  def _strong_scaling_lines(self,data,axis):
    totals=data[1]
    ref=totals[0] # the reference value is assumed to be the first
    coreref=float(self.ncores[0])
    #print self.totals
    if self.vals=='Seconds':
      label='Speedup'
      dataline=[ref/b for b in totals]
    else:
      label='Parallel Efficiency'
      dataline=[ (ref/b)/(c/coreref)  for b,c in zip(totals,self.ncores)]
    self.draw_lineplot(axis.twin,dataline,label)


  def draw_barfigure(self,fig,axis,data,title):
    import numpy as np
    from matplotlib.widgets import RadioButtons,Button
    import matplotlib.pyplot as plt
    from futile import Figures as Fig
    if self.static: fig.patch.set_facecolor("white")
    #write the plot in the axis
    self._draw_barplot(axis,data[0],self.vals,title=title,nokey=self.nokey)
    #add the lines useful for a strong scaling test
    if self.strong_scaling: self._strong_scaling_lines(data,axis)
    #if self.vals == 'Percent': axis.set_yticks(np.arange(0,100,10))
    #then add buttons to the plot
    if self.radio is None and not self.static:
      try:
          self.radio = RadioButtons(plt.axes([0.0, 0.75, 0.08, 0.11], axisbg='lightgoldenrodyellow'), ('Seconds', 'Percent'),
                                    active=1 if self.vals=='Percent' else 0)
      except AttributeError:
          self.radio = RadioButtons(plt.axes([0.0, 0.75, 0.08, 0.11], facecolor='lightgoldenrodyellow'), ('Seconds', 'Percent'),
                                    active=1 if self.vals=='Percent' else 0)
      self.radio.on_clicked(self.replot)
      tt=axis.get_xticks()
      routics=[Fig.axis_from_data(fig,axis,(tic-0.45*self.barwidth,self.barwidth)) for tic in tt]
      unbtics=[Fig.axis_from_data(fig,axis,(tic+0.15*self.barwidth,self.barwidth)) for tic in tt]
      self.routine_buttons=[]
      from functools import partial
      for i,t in enumerate(routics):
        if self.routines[i] is None: continue
        but=Button(plt.axes([t[0], 0.0, 0.15*t[1], 0.05]), 'R')
        but.on_clicked(partial(self.routines_plot,i))
        self.routine_buttons.append(but)
      for i,t in enumerate(unbtics):
        if not self.unbalancings[i]: continue
        but=Button(plt.axes([t[0], 0.0, 0.15*t[1], 0.05]), 'U')
        but.on_clicked(partial(self.workload_plot,i))
        self.routine_buttons.append(but)
      #fig.canvas.mpl_connect('pick_event',self._onclick_ev)
      fig.canvas.mpl_connect('button_press_event',
                             partial(self._onclick_ev,fig,axis))
    fig.canvas.draw()

  def routines_plot(self,index,event=None):
    "Draw the plot of the routines level for the run identified by index"
    toplt=self.routines[index]
    title='Routines '+str(self.ids[index])+'-'+str(index)
    if self.figures.exists(title): return
    #data=dump_timing_level(toplt)#,starting_point='cluster')
    fig,ax=self.figures.add(title=title,
                            axes={'rect':[0.025,0.025,0.95,0.95],
                                  'polar':True})
    if not hasattr(self,'routine_plots'): self.routine_plots=[]
    plt=polar_axis(fig,ax,toplt)
    self.routine_plots.append(plt)
    fig.show()

  def workload_plot(self,index,event=None):
    "Draw the plot of the workload of different classes for the run identified by index"
    from functools import partial
    title='Unbalancing '+str(self.ids[index])+'-'+str(index)
    if self.figures.exists(title): return
    fig,ax=self.figures.add(title=title)
    leg,plts,show=self.load_unbalancing(ax,self.scf[index]["Classes"],self.hostnames[index])
    idx=len(self.lined)
    self.lined.append({})
    for legline, origline,yes in zip(leg.get_patches(), plts, show):
          legline.set_picker(5)  # 5 pts tolerance
          self.lined[idx][legline] = origline
          if not yes: self._toggle_visible(origline,legline)
    fig.canvas.mpl_connect('pick_event', partial(self._onpick_curve,idx,fig))
    fig.show()

  def find_items(self,category,dict_list):
    """For a given category find the items which have them"""
    import numpy as np
    items={}
    for idoc in range( len(dict_list) ):
        for cat in dict_list[idoc]["Categories"]:
            dicat=dict_list[idoc]["Categories"][cat]
            if dicat["Class"] == category:
                if cat not in items:
                    items[cat]=np.zeros(len(dict_list))
                items[cat][idoc]=dicat["Data"][self.iprc]
    return [ (cat,items[cat]) for cat in items],[ doc["Classes"][category][1] for doc in dict_list]

  def collect_categories(self,dict_list,vals):
    """ Collect all the categories which belong to all the dictionaries """
    import numpy as np
    if vals == 'Percent':
      self.iprc=0
    elif vals == 'Seconds':
      self.iprc=1
    catsdats=[]
    self.values_legend=[]
    totals=[]
    for cat in self.classes:
      try:
        if cat == "Unknown": #this is always done
            data_unk=[]
            for doc in dict_list:
                totals.append(doc["Classes"]["Total"][1])
                percent_unk=100.0-doc["Classes"]["Total"][0]
                if self.iprc==0:
                    data_unk.append(percent_unk)
                elif self.iprc==1:
                    time_unk=(doc["Classes"]["Total"][1])*percent_unk/100.0
                    data_unk.append(time_unk)
            dat=np.array(data_unk)
        else:
            dat=np.array([doc["Classes"][cat][self.iprc] for doc in dict_list])
        catsdats.append((cat,dat))
        self.values_legend.append(cat)
      except Exception as e:
        print('EXCEPTION FOUND',e)
        print("category",cat,"not present everywhere")
    return catsdats,totals

  def _onclick_ev(self,fig,axis,event):
    import matplotlib.pyplot as plt
    #thisline = event.artist
    #xdata, ydata = thisline.get_xy()
    #print dir(event)
    #print event.inaxes
    xdata,ydata=event.x,event.y #event.xdata,event.ydata
    if xdata is None or ydata is None: return
    #print 'picked',xdata,ydata
    xdata,ydata=axis.transData.inverted().transform((xdata,ydata))
    #print 'transformed',xdata,ydata
    #data_from_data(fig,dst=axis,src=axis.twin,data=(int(xdata),ydata))
    #print 'converted',xdata,ydata
    if ydata < 0.0 or xdata < 0.0 or xdata >= len(self.scf): return
    xdata=int(xdata)
    #find the category which has been identified
    y0data=0.0
    category=None
    for cat in self.values_legend:
      if cat == 'Unknown': continue
      y0data+=self.scf[xdata]["Classes"][cat][self.iprc]
      #print 'cat,y0data',cat,y0data,ydata
      if y0data > ydata:
        category=cat
        break
    #print 'category',category
    if category is not None and category != "Unknown": self.inspect_category(category)
  def _barplot_gnuplot(self,data,ids,lookup=None,select_category=None):
      import Gnuplot
      import numpy as np
      gpdata=[]
      for cat,dat in data:
        if select_category is not None and cat not in select_category: continue
        #gpdata.append(Gnuplot.Data(np.arange(len(dat)),dat,title=cat))
        lup=np.arange(len(dat)) if lookup is None else lookup
        toplt=[dat[0]]+dat[lup].tolist()
        gpdata.append(Gnuplot.Data(toplt,title=cat))
      names=[ids[i] for i in lup]
      return gpdata,names

  def _aggregate_names(self,data,list_agg):
    """
    Aggregate the names of the plot in some different categories
    the structure of dict_agg should ba a list of tuples
    (newkey ,[list of oldkeys])
    between the original keys and the desired keys
    """
    newdata={newkey: 0.0 for newkey,oldkeys in list_agg}
    for cat,dat in data:
      found=False
      for newkey,oldkeys in list_agg:
        if cat in oldkeys:
          newdata[newkey]+=dat
          found=True
          break
      if not found: newdata[cat]=dat
    result=[]
    for newkeys,oldkeys in list_agg:
      result.append((newkeys,newdata.pop(newkeys)))
    for newkeys in newdata:
      result.append((newkeys,newdata[newkeys]))
    return result

  def gnuplot_figure(self,lookup=None,epsfile=None,aggregate=None,select_category=None):
      """Create a figure to be plotted with gnuplot

      Create a gnuplot histogram that can be plotted for production results.

      Args:
         lookup (list): A list of the items of the `TimeData` instance that will be considered for the histogram
         epsfile (str): Name of the eps file in which the data will be plot
         aggregate (list): list of tuples of the categories to aggregate, accoding to the convention
              (newkey ,[list of oldkeys]) between the original keys and the desired keys
         select_category (str): Name of the category to be plot, if specified the internal items of the category will be plot.
      """
      import Gnuplot,numpy as np
      data=self.collect_categories(self.scf,self.vals)
      dataplot=data[0] if aggregate is None else self._aggregate_names(data[0],aggregate)
      gpdata,names=self._barplot_gnuplot(dataplot,self.ids,lookup,select_category)
      gp=Gnuplot.Gnuplot()
      gp("""
set style fill  pattern 1 border -1
set style rectangle back fc  lt -3 fillstyle  solid 1.00 border -1
set key title ""
set key outside right bottom vertical Left reverse enhanced autotitles columnhead nobox
set key invert samplen 4 spacing 1 width 0 height 0
set style increment default
unset style line
unset style arrow
set style histogram rowstacked title  offset character 0, 0, 0
set pointsize 1
set style data histograms
set style function lines
set ytics border in scale 1,0.5 nomirror norotate  offset character 0, 0, 0
set ytics autofreq
set y2tics border in scale 1,0.5 nomirror norotate  offset character 0, 0, 0
set y2tics autofreq
set xtics 0,0.5,0
""")
      gp('set xrange [-1:'+str(len(names))+']')
      if self.vals=='Percent': gp('set yrange [0:100]')
      #gp('set xtics add (" " -1)')
      for i,xtic in enumerate(names):
          gp('set xtics add ("'+str(xtic)+'" '+str(i)+') rotate by 45 right')
      #gp('set xtics add (" " '+str(len(names))+')')
      gp.ylabel(self.vals)
      gp.plot(*gpdata)
      if epsfile is not None: gp.hardcopy(filename=epsfile,eps=True,enhanced=True,color=True)
      return gp

  def _draw_barplot(self,axbars,data,vals,title='Time bar chart',static=False,nokey=False):
    import numpy as np
    from pylab import cm as cm
    ndata=len(data[0][1])
    ind=np.arange(ndata)
    bot=np.array(ndata*[0.0])
    width=self.barwidth
    icol=1.0
    for cat,dat in data:
      #print 'cat',cat,dat
      #for i in range(len(self.ids)):
      #  print self.ids[i],dat[i]
      plt=axbars.bar(ind,dat,width,bottom=bot,color=cm.jet(icol/len(data)),picker=True,label=cat)
      #self.plts.append(plt)
      bot+=dat
      icol+=1.0
    #drawn_classes=np.array(self.values_legend)
    axbars.set_title(title,fontsize=self.fontsize)
    axbars.set_ylabel(vals,fontsize=self.fontsize)
    axbars.set_xticks(ind+width/2.)
    axbars.set_xticklabels(np.array(self.ids),size=self.fontsize)
    if not nokey:
      #self.leg = axbars.legend(loc='upper right',fontsize=self.fontsize)
      self.leg = axbars.legend(loc='best',fontsize=self.fontsize)
      self.leg.get_frame().set_alpha(0.4)
    #treat the totals differently
    return bot

  def draw_lineplot(self,ax,data,label):
    #ax=axbars.twinx()
    import numpy as np
    ind=np.arange(len(data))+self.barwidth/2.
    ax.plot(ind,data,'o-',color='red',label=label)
    ax.set_ylabel(label,fontsize=self.fontsize)

  def replot(self,label):
    self.vals=label
    #print self.vals
    for i,fig in enumerate(self.figures):
      fx,ax=fig
      cat=ax.get_title()
      ax.cla()
      if i==0:
        data=self.collect_categories(self.scf,label)
      else:
        data=self.find_items(cat,self.scf)
      if data != []: self.draw_barfigure(fx,ax,data,cat)

  def unbalanced(self,val):
    """Criterion for unbalancing"""
    return val > 1.1 or val < 0.9

  def find_unbalanced(self,data):
    """Determine lookup array of unbalanced categories"""
    ipiv=[]
    for i in range(len(data)):
      if self.unbalanced(data[i]):
        ipiv.append(i)
    return ipiv

  def load_unbalancing(self,ax,dict,hosts):
    """Extract the data for plotting the hostname balancings between different categories in bar chart"""
    import numpy as np
    from pylab import cm
    width=0.50
    plts=[]
    key_legend=[]
    values_legend=[]
    icol=1.0
    show_initially=[]
    for cat in self.classes:
      if cat=='Unknown': continue
      try:
        unb=np.array(dict[cat][2:])
        #print 'unbalancing',unb
        #unb2=self.find_unbalanced(unb)
        #print 'unbalanced objects',cat
        #if hosts is not None and (cat=='Convolutions' or cat =='Communications'):
          #print 'unb2',unb2,len(unb),len(hosts)
          #print 'vals',[ [i,unb[i],hosts[i]] for i in unb2]
        ind=np.arange(len(unb))
        pltmp=ax.bar(ind,unb,width,color=cm.jet(icol/len(self.classes)))
        plts.append(pltmp)
        key_legend.append(pltmp[0])
        values_legend.append(cat)
        show_initially.append(dict[cat][0] > 5.0)
        icol+=1.0
        if (width > 0.05):
          width -= 0.05
      except Exception as e:
        print('EXCEPTION FOUND',e)
        print("cat",cat,"not found in workload data")
        return None

    if len(ind) > 2:
      tmp=np.array(hosts) if hosts is not None else None
    else:
      tmp=np.array(["max","min"])
    ax.set_ylabel('Load Unbalancing wrt average')
    ax.set_title('Work Load of different classes')
    if tmp is not None:
      ax.set_xticks(ind+width/2.)
      ax.set_xticklabels(tmp,rotation=90,verticalalignment='bottom')
    ax.set_yticks(np.arange(0,2,0.25))
    ax.axhline(1.0,color='k',linestyle='--')
    leg=ax.legend(np.array(key_legend),np.array(values_legend))#,fancybox=True, shadow=True)
    #leg.get_frame().set_alpha(0.4)
    #show only the most relevant categories
    return leg,plts,show_initially

  def _onpick_curve(self,index,fig,event):
        # on the pick event, find the orig line corresponding to the
    # legend proxy line, and toggle the visibility
    legline = event.artist
    origline = self.lined[index][legline]
    self._toggle_visible(origline,legline)
    fig.canvas.draw()

  def _toggle_visible(self,data,legend):
    for val in data:
      vis = not val.get_visible()
      val.set_visible(vis)
    #last value is like all the others
    if vis:
        legend.set_alpha(1.0)
    else:
        legend.set_alpha(0.2)



"""Actions to define on the Input parameters.

This module defines some of the most common actions that a BigDFT user might
like to perform on the input file. Such module therefore sets some of the keys
of the input dictionary to the values needed to perform the operations.
Users might also inspire to the actions performed in order to customize the
runs in a different way. All the functions of this module have as first
argument ``inp``, the dictionary of the input parameters.

Many other actions are available in BigDFT code. This module only regroups the
most common. Any of these functionalities might be removed from the input file
by the :py:func:`remove` function.

Note:

   Any of the action of this module, including the :py:func:`remove` function,
   can be also applied to an instance of the
   :py:class:`BigDFT.Inputfiles.Inputfile` class, by removing the first
   argument (``inp``). This adds extra flexibility as the same method may be
   used to a dictionary instance or to a BigDFT input files.
   See the example :ref:`input_action_example`.

Note:

   Each of the actions here **must** have default value for the arguments
   (except the input dictionary ``inp``). This is needed for a good behaviour
   of the function `remove`.


.. autosummary::

   remove
   set_xc
   set_hgrid
   set_rmult
   set_wavefunction_convergence
   set_atomic_positions
   set_mesh_sizes
   optimize_geometry
   spin_polarize
   charge
   charge_and_polarize
   set_symmetry
   apply_electric_field
   set_random_inputguess
   write_orbitals_on_disk
   read_orbitals_from_disk
   write_density_on_disk
   calculate_dipole
   use_gpu_acceleration
   change_data_directory
   connect_run_data
   add_empty_SCF_orbitals
   extract_virtual_states
   set_electronic_temperature
   calculate_tddft_coupling_matrix
   write_support_function_matrices

"""

from futile.Utils import dict_set

__set__ = dict_set
"""func: Action function.

This is the pointer to the set function, useful to modify the action with the
undo method

"""


def __undo__(inp, *subfields):
    """
    Eliminate the last item of the subfields as provided to dict_set
    """
    from futile.Utils import push_path
    # remove the last key until the parent is empty
    lastkey = -1
    tmp = {}
    while len(subfields) > -lastkey and tmp == {}:
        keys = subfields[:lastkey]
        tmp, k = push_path(inp, *keys)
        tmp.pop(k)
        lastkey -= 1


def remove(inp, action):
    """Remove action from the input dictionary.

    Remove an action from the input file, thereby restoring the **default**
    value, as if the action were not specified.

    Args:
       inp (dict): dictionary to remove the action from.
       action (func): one of the actions of this module. It does not need to be
       specified before, in which case it produces no effect.

    Example:
       >>> from Calculators import SystemCalculator as C
       >>> code=C()
       >>> inp={}
       >>> set_xc(inp,'PBE')
       >>> write_orbitals_on_disk(inp)
       >>> log=code.run(input=inp) # perform calculations
       >>> remove(write_orbitals_on_disk) #remove the action
       >>> read_orbitals_from_disk(inp)
       >>> # this will restart the SCF from the previous orbitals
       >>> log2=code.run(input=inp)
    """
    global __set__
    __set__ = __undo__
    action(inp)
    __set__ = dict_set


def set_hgrid(inp, hgrids=0.4):
    """
    Set the wavelet grid spacing.

    Args:
       hgrid (float,list): list of the grid spacings in the three directions.
       It might also be a scalar, which implies the same spacing
    """
    __set__(inp, 'dft', 'hgrids', hgrids)


def set_wavefunction_convergence(inp, gnrm=1.0e-04):
    """
    Set the tolerance acceptance level for stopping the self-consistent
    iterations

    Args:
       gnrm (float): the tolerance level
    """
    __set__(inp, 'dft', 'gnrm_cv', gnrm)


def set_rmult(inp, rmult=None, coarse=5.0, fine=8.0):
    """
    Set the wavelet grid extension by modifying the multiplicative radii.

    Args:
       rmult (float,list): list of two values that have to be used for the
         coarse and the fine resolution grid. It may also be a scalar.
       coarse (float): if the argument ``rmult`` is not provided it sets the
         coarse radius multiplier
       fine (float): if the argument ``rmult`` is not provided it sets the fine
         radius multiplier
    """
    rmlt = [coarse, fine] if rmult is None else rmult
    __set__(inp, 'dft', 'rmult', rmlt)


def set_symmetry(inp, yes=True):
    """
    Set the symmetry detection for the charge density and the ionic forces and
    stressdef set_symmetry(inp,yes=True):

    Args:
       yes (bool): If ``False`` the symmetry detection is disabled
    """
    __set__(inp, 'dft', 'disablesym', not yes)


def set_linear_scaling(inp):
    """
    Activates the linear scaling mode
    """
    newid = 'linear'
    previous_ipid = inp.get('dft', 'False')
    if previous_ipid:
        previous_ipid = inp.get('inputpsiid', 'False')
    if previous_ipid == 2:
        newid = 102

    __set__(inp, 'dft', 'inputpsiid', newid)


def set_mesh_sizes(inp, ngrids=64):
    """
    Constrain the number of grid points in each direction.
    This is useful when performing periodic system calculations with variable
    cells which need to be compared each other. In this way the number of
    degrees of freedom is kept constant throughout the various simuilations.

    Args:
       ngrids (int,list): list of the number of mesh points in each direction.
         Might be a scalar.
    """
    __set__(inp, 'dft', 'ngrids', ngrids)


def spin_polarize(inp, mpol=1):
    """
    Add a collinear spin polarization to the system.

    Arguments:
       mpol (int): spin polarization in Bohr magneton units.
    """
    __set__(inp, 'dft', 'nspin', 2)
    __set__(inp, 'dft', 'mpol', mpol)


def charge(inp, charge=-1):
    """
    Charge the system

    Arguments:
        charge (int,float): value of the charge in units of *e* (the electron
          has charge -1). Also accept floating point numbers.
    """
    __set__(inp, 'dft', 'qcharge', charge)


def apply_electric_field(inp, elecfield=[0, 0, 1.e-3]):
    """
    Apply an external electric field on the system

    Args:
       electric (list, float): Values of the Electric Field in the three
         directions. Might also be a scalar.
    """
    __set__(inp, 'dft', 'elecfield', [e for e in elecfield])


def charge_and_polarize(inp):
    """
    Charge the system by removing one electron. Assume that the original
      system is closed shell, thus polarize.
    """
    charge(inp, charge=1)
    spin_polarize(inp, mpol=1)


def set_SCF_method(inp, method='dirmin', mixing_on='density',
                   mixing_scheme='Pulay'):
    """
    Set the algorithm for SCF.

    Args:
       method (str): The algoritm chosen. Might be different for the cubic (CS)
         or linear scaling (LS) algorithm.
         * dirmin: Direct minimization approach (valid both to LS and CS)
         * mixing: Mixing scheme (only CS)
         * foe: Fermi Operator Expansion (only LS)
         * pexsi: Pole EXpansion and Selected Inversion method (only LS,
             require PEXSI compilation)
         * diag: Explicit diagonalization (only LS, for validation purposes)

       mixing_on (str): May be ``"density"`` or ``"potential"`` in the
         ``"mixing"`` case, decide to which quantity the mixing to be performed

       mixing_scheme (str): May be:

          * Pulay : DIIS mixing on the last 7 iterations

          * Simple: Simple mixing

          * Anderson: Anderson scheme

          * Anderson2: Anderson scheme based on the two pervious iterations

          * CG: Conjugate Gradient based on the minimum of the energy with
                respect of the potential

    Warning:
       Only the FOE method exhibit asymptotic linear scaling regime.

    Todo:
       Check if the linear scaling case needs another input variable for the
       mixing of the potential (density)

    """
    method.upper()
    if method != 'MIXING':
        __set__(inp, 'lin_kernel', 'linear_method', method)
    if method == 'DIRMIN':
        __set__(inp, 'mix', 'iscf', 0)
        return
    iscf = 0
    if mixing_on == 'density':
        iscf += 10
    if mixing_scheme == 'Pulay':
        iscf += 7
    if mixing_scheme == 'Anderson':
        iscf += 3
    if mixing_scheme == 'Anderson2':
        iscf += 4
    if mixing_scheme == 'Simple':
        iscf += 2
    if mixing_scheme == 'CG':
        iscf += 5
    __set__(inp, 'mix', 'iscf', iscf)


def add_empty_SCF_orbitals(inp, norbs=10):
    """
    Insert ``norbs`` empty orbitals in the SCF procedure

    Args:
       norbs (int): Number of empty orbitals

    Warning:
       In linear scaling case, this is only meaningful for the direct
       minimization approach.
    """
    __set__(inp, 'mix', 'norbsempty', norbs)
    __set__(inp, 'lin_general', 'extra_states', norbs)


def write_cubefiles_around_fermi_level(inp, nplot=1):
    """
    Writes the ``nplot`` orbitals around the fermi level in cube format

    Args:
      nplot (int): the number of orbitals to print around the fermi level.

    Warning:
       This is presently meaningful only for a empty states calculation.

    Warning:
       This would work only for the cubic scaling code at present.
    """
    __set__(inp, 'dft', 'nplot', nplot)


def write_orbitals_on_disk(inp, format='binary'):
    """
    Set the code to write the orbitals on disk in the provided format

    Args:
      format (str): The format to write the orbitals with. Accepts the strings:
         * 'binary'
         * 'text'
         * 'etsf' (requires etsf-io enabled)

    Todo:
      Verify if this option works for a linear scaling calulation.
    """
    fmt = format
    __set__(inp, 'output', 'orbitals', fmt)


def write_support_functions_on_disk(inp, format='binary', matrices=True,
                                    coefficients=False):
    pass


def write_support_function_matrices(inp, format='text'):
    """
    Write the matrices of the linear scaling formats.

    Args:
       format (str): The format to write the orbitals with. Accepts the
         strings:
          * 'binary'
          * 'text'

    Todo:
       Verify if the binary format is available and set the appropriate values
    """
    fmt = 0
    if format == 'text':
        fmt = 1
    elif format == 'binary':
        fmt = 4
    __set__(inp, 'lin_general', 'output_mat', fmt)


def set_atomic_positions(inp, posinp=None):
    """
    Insert the atomic positions as a part of the input dictionary
    """
    __set__(inp, 'posinp', posinp)


def read_orbitals_from_disk(inp):
    """
    Read the orbitals from data directory, if available
    """
    newid = 2
    previous_ipid = inp.get('dft', 'False')
    if previous_ipid:
        previous_ipid = inp.get('inputpsiid', 'False')
    if previous_ipid == 'linear' or previous_ipid == 100:
        newid = 102
    __set__(inp, 'dft', 'inputpsiid', newid)


def set_random_inputguess(inp):
    """
    Input orbitals are initialized as random coefficients
    """
    __set__(inp, 'dft', 'inputpsiid', -2)


def set_electronic_temperature(inp, kT=1.e-3, T=0):
    """
    Define the electronic temperature, in AU (``kT``) or K (``T``)
    """
    TtokT = 8.617343e-5/27.21138505
    tel = TtokT*T if T != 0 else kT
    __set__(inp, 'mix', 'tel', tel)


def optimize_geometry(inp, method='FIRE', nsteps=50):
    """
    Optimize the geometry of the system

    Args:
       nsteps (int): maximum number of atomic steps.
       method (str): Geometry optimizer. Available keys:
          * SDCG:   A combination of Steepest Descent and Conjugate Gradient
          * VSSD:   Variable Stepsize Steepest Descent method
          * LBFGS:  Limited-memory BFGS
          * BFGS:   Broyden-Fletcher-Goldfarb-Shanno
          * PBFGS:  Same as BFGS with an initial Hessian obtained from a force
                    field
          * DIIS:   Direct inversion of iterative subspace
          * FIRE:   Fast Inertial Relaxation Engine as described by Bitzek et
                    al.
          * SBFGS:  SQNM minimizer, keyword deprecated, will be replaced by
                    SQNM in future release
          * SQNM:   Stabilized quasi-Newton minimzer
    """
    __set__(inp, 'geopt', 'method', method)
    __set__(inp, 'geopt', 'ncount_cluster_x', nsteps)


def set_xc(inp, xc='PBE'):
    """
    Set the exchange and correlation approximation

    Args:
       xc (str): the Acronym of the XC approximation

    Todo:
       Insert the XC codes corresponding to ``libXC`` conventions
    """
    __set__(inp, 'dft', 'ixc', xc)


def write_density_on_disk(inp):
    """
    Write the charge density on the disk after the last SCF convergence
    """
    __set__(inp, 'dft', 'output_denspot', 21)


def use_gpu_acceleration(inp):
    """
    Employ gpu acceleration when available, for convolutions and Fock operator

    Todo:
       Verify what happens when only one of the functionality is enabled at
       compile-time
    """
    __set__(inp, 'perf', 'accel', 'OCLGPU')
    __set__(inp, 'psolver', 'setup', 'accel', 'CUDA')


def set_wavefunction_iterations(inp, nit=[50, 1]):
    """
    Set the number of the iteration per loop

    Args:
       nit (int,list): integer of the number of iterations. Might be a scalar
            or a list, up to length two. The first element of the list contains
            the number of iterations of the direct minimization loop. if
            ``nit`` is a scalar, only this contribution is taken into account.
            The second element is the number of subspace iterations where the
            hamiltonian is diagonalized in the subspace.
    """
    try:
        nlen = len(nit)
    except TypeError:
        nlen = 0
    if nlen >= 1:
        __set__(inp, 'dft', 'itermax', nit[0])
    if nlen == 2:
        __set__(inp, 'dft', 'nrepmax', nit[1])
    if nlen == 0:
        __set__(inp, 'dft', 'itermax', nit)


def change_data_directory(inp, name=''):
    """
    Modify the name of the ``data-`` directory.
    Useful to grab the orbitals from another directory than the run name

    Args:
       name (str): the name of the run
    """
    __set__(inp, 'radical', name)


def calculate_tddft_coupling_matrix(inp, tda=False, rpa=True, fxc=True):
    """
    Perform a Casida TDDFT coupling matrix extraction.

    Args:
       tda (bool): when ``True``, Tamm-Dancoff approximation is used for the
         extraction of the coupling matrix
       rpa (bool): when ``False``, the calculation of the RPA term (the linear
         response of the hartree potential) is switched off
       fxc (bool): when ``False``, the calculation of the fxc term (the linear
         response of the XC operator) is switched off.

    Note:
       The arguments ``fxc`` and ``rpa`` should not be simultaneously
       ``False``.

    Warning:
       Presently the LR-TDDFT casida fxc is only available for LDA
       functionals in ABINIT flavour.
    """
    approach = 'TDA' if tda else 'full'
    __set__(inp, 'tddft', 'tddft_approach', approach)
    if rpa and fxc:
        output = 'complete'
    elif rpa:
        output = 'rpa'
    else:
        output = 'fxc'
    __set__(inp, 'output', 'coupling_matrix', output)


def extract_virtual_states(inp, nvirt, davidson=False):
    """
    Extract a given number of empty states **after** the scf cycle.

    Args:
       davidson (bool): If set to ``True`` activates davidson calculation,
       otherwise Trace Minimization of the Hamiltonian is employed.
    """
    nv = nvirt if davidson else -nvirt
    __set__(inp, 'dft', 'norbv', nv)
    __set__(inp, 'dft', 'nvirt', nvirt)
    __set__(inp, 'dft', 'itermax_virt', 150)


def connect_run_data(inp, log=None):
    """
    Associate the data of the run of a given logfile to the input
    by retrieving the data directory name of the logfile.

    Args:
       log (Logfile): instance of a Logfile class

    """
    if log is None:
        change_data_directory(inp)  # no effect
    else:
        ll = log if len(log) == 0 else log[0]
        change_data_directory(inp, ll.log['radical'])


def calculate_dipole(inp):
    """
    Extract the dipole momenet from the total charge density.

    Note:
      This function is useful for the linear scaling setup as the cubic
      scaling approach always calculates the charge density multipoles.
    """
    __set__(inp, 'lin_general', 'calc_dipole', True)



'''
This module defines the atom class, which is a class which contains very
general descriptions of a single atom.

Atoms act as dictionary types. Allowed dictionary values are defined in
the ``MULTIPOLE_ANALYSIS_KEYS`` and ``PROTECTED_KEYS`` lists.
'''
try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
from futile.Utils import write as safe_print

#: Conversion between Atomic Units and Bohr
AU_to_A = 0.52917721092
#: A list of valid keys for describing a multipole.
MULTIPOLE_ANALYSIS_KEYS = ['q0', 'q1', 'q2',
                           'sigma', 'multipole character']
#: A list of valid keys for describing an atom.
PROTECTED_KEYS = MULTIPOLE_ANALYSIS_KEYS + \
    ["frag"] + ["r", "units", 'force', 'velocity']


def nzion(sym):
    """
    Returns the charge of the nucleus associated with an atomic symbol.

    Args:
      sym (str): the atomic symbol of this atom.

    Returns:
      (float): the charge of that nucleus.
    """
    return _nzion[sym]


def atomic_number(sym):
    """
    Returns the atomic number associated with the given symbol.

    Args:
      sym (str): the atomic symbol of this atom.

    Returns:
      (int): the atomic number
    """
    return _atomic_number[sym]


def number_to_symbol(number):
    """
    Returns the symbol of atoms with a given atomic number.

    Args:
      number (int): the atomic number to lookup.

    Returns:
      (str): the atomic symbol with the given number.
    """
    return [x for x in _atomic_number if _atomic_number[x] == number][0]


_atomic_number = {"H": 1, "He": 2,
                  "Li": 3, "Be": 4, "B": 5, "C": 6,
                  "N": 7, "O": 8, "F": 9, "Ne": 10,
                  "Na": 11, "Mg": 12, "Al": 13, "Si": 14,
                  "P": 15, "S": 16, "Cl": 17, "Ar": 18,
                  "Cu": 29, "Zn": 30}


_nzion = {"H":  1.0, "He": 2.0,
          "Li": 1.0, "Be": 2.0, "B":  3.0, "C":  4.0,
          "N": 5.0, "O": 6.0, "F":  7.0, "Ne": 8.0,
          "Na": 1.0, "Mg": 2.0, "Al": 3.0, "Si": 4.0,
          "P":  5.0, "S":  6.0, "Cl": 7.0, "Ar": 8.0,
          "Cu": 11.0}


class Atom(MutableMapping):
    """
    Defines a wrapper for atoms.

    An atom may have many quantities associated with it. These quantities
    are get and set in a dictionary like fashion, allowing an atom to
    dynamically hold whatever data you need. However, we still wrap it in a
    class so that we can have some common operations for it, as well as so we
    can maintain suitable units.

    It is this class's responsibility to extract the main properties of an
    atom (position, symbol) from the dictionary.

    Args:
      data (dict):
        A dictionary of miscellaneous values to associate with this atom.
    """

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))

    @property
    def atomic_number(self):
        """
        The atomic number of this atom.
        """
        return atomic_number(self.sym)

    def dict(self):
        """
        Convert to a dictionary.
        """
        return self.store

    def get_external_potential(self, units="bohr"):
        """
        Transform the atom into a dictionary ready to be put as external
        potential.
        """
        from numpy import ndarray
        return_dict = {}
        return_dict["sym"] = self.sym
        return_dict["r"] = self.get_position(units)
        for k in MULTIPOLE_ANALYSIS_KEYS:
            if k in self:
                val = self[k]
                if isinstance(val, ndarray):
                    return_dict[k] = list(val)
                else:
                    return_dict[k] = val

        return return_dict

    @property
    def nel(self):
        """
        The number of electrons in this atom.
        """
        return nzion(self.sym)

    @property
    def q0(self):
        """
        Provides the charge of the atom.
        """
        charge = self.get('q0')
        if charge is not None:
            charge = charge[0]
        return charge

    @property
    def q1(self):
        """
        Provides the dipole of the atom.
        """
        import numpy as np
        dipole = self.get('q1')  # they are (so far) always given in AU
        if dipole is not None:
            dipole = np.array([dipole[2], dipole[0], dipole[1]])
        return dipole

    def set_multipole(self, mp, correct_charge=True):
        """
        Given another atom or a dictionary, this sets the multipole related
        values of this with those values.

        Todo:
          Arrive at a standard that avoids having to do the charge
          correction here.

        Args:
          mp (dict): a dictionary which contains information about multipoles.
          correct_charge (bool): currently there is an inconsistency in
            terms of gross charge, and this corrects it.
        """
        from copy import deepcopy

        for key in MULTIPOLE_ANALYSIS_KEYS:
            if key in mp:
                self[key] = deepcopy(mp[key])

        # Correct the charge
        if correct_charge and "q0" in self and "nzion" in mp:
            self["q0"][0] += mp["nzion"]
            self['multipole character'] = 'net'

    def get_force(self):
        """
        Returns the force on the atom in the desired units.

        Returns:
          An array of position values.
        """
        return self.get("force")

    def set_force(self, force):
        """
        Given an atom or a dictionary, this sets the force.

        Args:
          force (list): a list of force values.
        """
        self.store["force"] = force

    @property
    def sym(self):
        sym = _GetSymbol(self.store)
        if sym == 'r':
            sym = self.store['sym']
        return sym

    @sym.setter
    def sym(self, v):
        if 'sym' in self.store:
            self.store['sym'] = v
        else:
            sym = self.sym
            val = self.store[sym]
            del self[sym]
            self[v] = val

    def get_position(self, units="bohr"):
        """
        Returns the position of the atom in the desired units.

        Args:
          units (str): the units to return the position in. Default is bohr.

        Returns:
          An array of position values.
        """
        from numpy import array

        # Grab the position from the store
        if 'r' in self.store:
            pos = self.store['r']
        else:
            pos = self.store[self.sym]
        pos = array([float(x) for x in pos])

        # Make sure the units are correct
        if _IsAngstroem(self):
            pos /= AU_to_A
        if _IsAngstroem(units):
            pos *= AU_to_A

        return [float(x) for x in pos]

    def set_position(self, new_pos, units="bohr"):
        """
        Set the position of the atom.

        Args:
          new_pos (list): a list of floats defining the new position.
          units(str): the units of the new position being passed. Default is
            bohr.
        """
        from numpy import array
        # Convert the input to the right units.
        pos = array(new_pos)
        if _IsAngstroem(units):
            pos /= AU_to_A
        if _IsAngstroem(self):
            pos *= AU_to_A
        pos = [x for x in pos]

        # Insert
        if 'r' in self.store:
            self.store['r'] = pos
        else:
            self.store[self.sym] = pos
        pass

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        if key == self.sym:
            raise ValueError("You can't delete the symbol from an atom.")
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key

    def __eq__(self, other):
        """
        Compare two atoms. They are equal if they have the same position and
        symbol.

        other (dict, Atom): the atom (or something that can be cast to one)
          to compare with.
        """
        from numpy.linalg import norm
        from numpy import array

        # Upcast to an Atom
        othercomp = Atom(other)

        # Compare Symbols
        sym1 = self.sym
        sym2 = othercomp.sym
        if sym1 != sym2:
            return False

        # Compare position
        pos1 = array(self.get_position())
        pos2 = array(othercomp.get_position())

        return norm(pos1 - pos2) < 1e-3


def _GetSymbol(atom):
    """
    Provides the key which defines the element of the of atom.

    Arguments:
      atom (dict): a dictionary describing the atom.
    Returns:
      (str): the symbol the atom.
    """
    ks = atom.keys()
    if 'sym' in ks:
        return atom['sym']

    for k in ks:
        if k not in PROTECTED_KEYS and isinstance(atom[k], list):
            if len(atom[k]) == 3:
                return k

    raise ValueError


def _IsAngstroem(units):
    """
    Checks if a string or atom has angstroem as its units.

    Args:
      units: either a string, or an ``Atom``.
    """
    if isinstance(units, Atom):
        check = units.store.get("units")
        if not check:
            return False
    else:
        check = units
    return check == "angstroem" or check == "angstroemd0"


if __name__ == "__main__":
    """Test the atom module"""
    safe_print("Access the full data")
    test_atom = Atom({'r': [1.0, 0.0, 0.0], 'sym': "He", 'units': 'bohr'})
    safe_print(dict(test_atom))
    # Access the derived data
    safe_print(test_atom.sym)
    safe_print(test_atom.get_position())
    safe_print(test_atom.get_position('angstroem'))
    safe_print()

    safe_print("Create a new atom with different units")
    new_atom = Atom({
        'r': [float(x) for x in test_atom.get_position('angstroem')],
        'sym': test_atom.sym, 'units': 'angstroem'})
    safe_print("Are these atoms equal?")
    safe_print(new_atom == test_atom)
    safe_print()

    safe_print("Now other times we get an array that looks more like this")
    test_atom = Atom(He=[1.0, 0.0, 0.0], units='bohr')
    safe_print(dict(test_atom))
    safe_print("But everything else works as expected")
    safe_print(test_atom.sym)
    safe_print(test_atom.get_position())
    safe_print(new_atom == test_atom)
    safe_print()

    safe_print("The atom can be used as a dict for adding new properties.")
    test_atom["frag"] = "ANA"
    for key, value in test_atom.items():
        safe_print(key, value)
    safe_print()
    safe_print("And if we update the dictionary position or symbol,")
    safe_print("everything else reacts with suitable caution.")
    test_atom["He"] = [-1.0, 0.0, 0.0]
    safe_print(dict(test_atom))
    safe_print(test_atom.get_position('angstroem'))

    safe_print("There is a protection to prevent you from deleting the")
    safe_print("symbol of an atom from the dictionary.")
    try:
        del test_atom["frag"]
        del test_atom["He"]
    except ValueError as v:
        safe_print(v)
    safe_print(dict(test_atom))
    safe_print()

    safe_print("But you can change the symbol if you are working with the")
    safe_print("other representation.")
    safe_print(dict(new_atom))
    new_atom["sym"] = "Na"
    safe_print(new_atom.sym)
    safe_print(dict(new_atom))
    safe_print()

    safe_print("One final check of the atom comparison")
    new_atom["units"] = "bohr"
    new_atom["r"] = [-1.0, 0.0, 0.0]
    safe_print(new_atom.sym, new_atom.get_position())
    safe_print(test_atom.sym, test_atom.get_position())
    safe_print(new_atom == test_atom)
    safe_print()

    safe_print("We can also update the position")
    safe_print(test_atom.get_position())
    safe_print(dict(test_atom))
    test_atom.set_position([1.0, 1.0, 1.0], units="angstroem")
    safe_print(test_atom.get_position(units="angstroem"))
    safe_print(dict(test_atom))
    safe_print()



""" Helper functions for automatically setting up computer & code.

Helper functions for setting up 

 1. An AiiDA localhost computer
 2. A "bigdft" code on localhost
 
Note: Point 2 is made possible by the fact that the ``bigdft`` executable is
available in the PATH on almost any UNIX system.
"""
from __future__ import absolute_import
from __future__ import print_function
import tempfile

LOCALHOST_NAME = 'localhost-test'

executables = {
    'bigdft': 'bigdft',
}


def get_path_to_executable(executable):
    """ Get path to local executable.

    :param executable: Name of executable in the $PATH variable
    :type executable: str

    :return: path to executable
    :rtype: str
    """
    # pylint issue https://github.com/PyCQA/pylint/issues/73
    import distutils.spawn  # pylint: disable=no-name-in-module,import-error
    path = distutils.spawn.find_executable(executable)
    if path is None:
        raise ValueError("{} executable not found in PATH.".format(executable))

    return path


def get_computer(name=LOCALHOST_NAME, workdir=None):
    """Get AiiDA computer.

    Loads computer 'name' from the database, if exists.
    Sets up local computer 'name', if it isn't found in the DB.
    
    :param name: Name of computer to load or set up.
    :param workdir: path to work directory 
        Used only when creating a new computer.

    :return: The computer node 
    :rtype: :py:class:`aiida.orm.Computer` 
    """
    from aiida.orm import Computer
    from aiida.common.exceptions import NotExistent

    try:
        computer = Computer.objects.get(name=name)
    except NotExistent:
        if workdir is None:
            workdir = tempfile.mkdtemp()

        computer = Computer(
            name=name,
            description='localhost computer set up by aiida_bigdft tests',
            hostname=name,
            workdir=workdir,
            transport_type='local',
            scheduler_type='direct')
        computer.store()
        computer.configure()

    return computer


def get_code(entry_point, computer):
    """Get local code.

    Sets up code for given entry point on given computer.
    
    :param entry_point: Entry point of calculation plugin
    :param computer: (local) AiiDA computer

    :return: The code node 
    :rtype: :py:class:`aiida.orm.Code` 
    """
    from aiida.orm import Code
    from aiida.common.exceptions import NotExistent

    try:
        executable = executables[entry_point]
    except KeyError:
        raise KeyError(
            "Entry point {} not recognized. Allowed values: {}".format(
                entry_point, list(executables.keys())))

    try:
        code = Code.get_from_string('{}@{}'.format(executable,
                                                   computer.get_name()))
    except NotExistent:
        path = get_path_to_executable(executable)
        code = Code(
            input_plugin_name=entry_point,
            remote_computer_exec=[computer, path],
        )
        code.label = executable
        code.store()

    return code



"""
This module provides helper subroutines for performing rototranslations.

In particular, this module can solve Wahba's problem, which seeks to compute
the optimal rotation matrix between two sets of points.

https://en.wikipedia.org/wiki/Kabsch_algorithm
"""


def apply_R(R, A):
    """
    Apply the rotation on the set of vectors A.

    Args:
      R (numpy.matrix): a rotation matrix.
      A (numpy.matrix): the set of vectors to rotate.

    Returns:
      (numpy.matrix): the set of vectors, rotated by the matrix.
    """
    A2 = R*A.T
    A2 = A2.T
    return A2


def apply_t(t, A):
    """
    Apply a translation on the set of vectors A.

    Args:
      t (numpy.matrix): a translation matrix.
      A (numpy.matrix): the set of vectors to translate.

    Returns:
      (numpy.matrix): the set of vectors, translated by the matrix.
    """
    from numpy import tile
    n = A.shape[0]

    return A + tile(t, (1, n)).T


def apply_Rt(R, t, A):
    """
    Rotate the element and apply the translation on the rotated vector.

    Args:
      R (numpy.matrix): a rotation matrix.
      t (numpy.matrix): a translation matrix.
      A (numpy.matrix): the set of vectors to translate.

    Returns:
      (numpy.matrix): the set of vectors, rototranslated.
    """
    RA = apply_R(R, A)
    return apply_t(t, RA)


def rigid_transform_3D(A, B, verbose=False):
    """
    Find the transformation R and t such that R*A + t ~= B, with an error
    quantified by J.

    Args:
      A (numpy.matrix): an NX3 matrix of points.
      B (numpy.matrix): a second NX3 matrix of points.
      verbose (bool): whether to be verbose during the calculation.

    Returns:
      (numpy.matrix): 3x3 rotation matrix/
      (numpy.matrix): 3X1 column vector representing the translation.
      (float): a quantification of the error.
    """
    from futile.Utils import write as safe_print
    from numpy import transpose, mean, tile, multiply, sqrt
    from numpy.linalg import svd, det

    assert len(A) == len(B)

    N = A.shape[0]  # total points

    centroid_A = mean(A, axis=0)
    centroid_B = mean(B, axis=0)

    # print 'centre',centroid_A,centroid_B
    # centre the points
    AA = A - tile(centroid_A, (N, 1))
    BB = B - tile(centroid_B, (N, 1))

    # dot is matrix multiplication for array
    H = transpose(AA) * BB

    # print 'H',H

    U, S, Vt = svd(H)

    R = Vt.T * U.T

    # special reflection case
    if det(R) < 0:
        if verbose:
            safe_print("#Reflection detected")
        Vt[2, :] *= -1
        R = Vt.T * U.T

    t = -R*centroid_A.T + centroid_B.T

    # print t

    # identify also the accuracy of wahba
    A2 = R*A.T + tile(t, (1, N))
    A2 = A2.T

    # Find the error
    err = A2 - B

    err = multiply(err, err)
    err = sum(err)
    rmse = sqrt(err/N)

    return R, t, rmse


def interpolate_points(A, B, steps, extrapolation_steps=0):
    """
    Given a set of points A and B, this generates a list of point sets
    that interpolate between A and B in a specified number of steps.

    Args:
      A (numpy.matrix): an NX3 matrix of points.
      B (numpy.matrix): a second NX3 matrix of points.
      steps (int): the number of steps to take between A and B.
      extrapolation_steps (int): optionally, we can extrapolate a number of
        steps beyond B on the same trajectory.

    Returns:
      (list): a list of points interpolating between A and B including
      A and B.
    """
    from scipy.linalg import funm
    from numpy import mean
    from copy import deepcopy
    R, t, rmse = rigid_transform_3D(A, B)

    if steps < 0:
        raise ValueError("Steps must be greater than or equal to zero.")

    # Scale the rotation
    R = funm(R, lambda x: x**(1.0/(steps+1)))

    # Apply each rotation.
    point_list = [deepcopy(A)]
    for i in range(1, steps+extrapolation_steps+2):
        new_point = deepcopy(point_list[-1])

        centroid = mean(new_point, axis=0)
        centroid = centroid.T

        new_point = apply_t(-1.0*centroid, new_point)
        new_point = apply_R(R, new_point)
        new_point = apply_t(centroid, new_point)

        point_list.append(new_point)

    # Compute a new translation matrix between the last two points.
    R, t, rmse = rigid_transform_3D(point_list[steps+1], B)
    t = 1.0/(steps+1) * t

    # Apply each translation.
    for i in range(1, steps+extrapolation_steps+2):
        point_list[i] = apply_t(i*t, point_list[i])

    # In order to remove any noise from the extrapolation, we copy
    # the points B at the end of the interpolation.
    point_list[steps+1] = deepcopy(B)

    return point_list


# Test with random data
if __name__ == '__main__':
    from futile.Utils import write as safe_print
    from numpy import random, mat, tile, multiply, sqrt
    from numpy.linalg import svd, det

    # Random rotation and translation
    R = mat(random.rand(3, 3))
    t = mat(random.rand(3, 1))

    # make R a proper rotation matrix, force orthonormal
    U, S, Vt = svd(R)
    R = U*Vt

    # remove reflection
    if det(R) < 0:
        Vt[2, :] *= -1
        R = U*Vt

    # number of points
    n = 6

    A = mat(random.rand(n, 3))
    B = R*A.T + tile(t, (1, n))
    B = B.T

    # recover the transformation
    ret_R, ret_t, ret_rmse = rigid_transform_3D(A, B)

    A2 = (ret_R*A.T) + tile(ret_t, (1, n))
    A2 = A2.T

    # Find the error
    err = A2 - B

    err = multiply(err, err)
    err = sum(err)
    rmse = sqrt(err/n)

    safe_print("Points A")
    safe_print(A)
    safe_print("")

    safe_print("Points B")
    safe_print(B)
    safe_print("")

    safe_print("Rotation")
    safe_print(R)
    safe_print("")

    safe_print("Translation")
    safe_print(t)
    safe_print("")

    safe_print("RMSE:", rmse)
    safe_print("If RMSE is near zero, the function is correct!")



"""
A module which contains the routines needed for computing the spillage values.
"""

#: Conversion between Atomic Units and Bohr
AU_to_A = 0.52917721092


class MatrixMetadata:
    """
    This class contains the information stored in the sparse matrix
    metadata file.

    Args:
      filename (str): the name of the metadata file to initialize from.

    Attributes:
      atoms (list): a list of atoms with their positions and associated
        basis functions.
      matdim (int): the dimension of the matrix.
    """
    def __init__(self, filename):
        from BigDFT.Atom import Atom
        self.atoms = []
        symbol_lookup = []
        with open(filename, "r") as ifile:
            # Read the first line
            matinfo = next(ifile).split()
            self.matdim, natoms, ntypes = [int(x) for x in matinfo[:3]]
            # Units
            next(ifile)
            # skip geocode
            line = next(ifile)
            # skip shift?
            line = next(ifile)
            # get the symbol lookup information
            for i in range(0, ntypes):
                line = next(ifile).split()
                symbol_lookup.append(line[2])
            # Get the atom positions
            for i in range(0, natoms):
                line = next(ifile).split()
                adict = {}
                adict["sym"] = symbol_lookup[int(line[0]) - 1]
                adict["r"] = [float(x) for x in line[1:4]]
                # THIS IS BECAUSE THE METADATA FILE CURRENTLY PRINTS THE
                # WRONG UNITS!
                adict["units"] = "bohr"
                adict["indices"] = []
                self.atoms.append(Atom(adict))
            # Get the indices
            for i in range(0, self.matdim):
                line = next(ifile).split()
                self.atoms[int(line[0]) - 1]["indices"].append(i)

    def get_frag_indices(self, system):
        """
        Retrive the indices of the matrix associated with each fragment
        of a system.

        Args:
          system (Fragments.System): the set of fragments to get the indices
            of.

        Return:
          (dict): a mapping from fragment to indices.

        Return:
          (dict): a mapping from fragment to indices.
        """
        matching = system.compute_matching(self.atoms)

        frag_indices = {}
        for fragid in system:
            frag_indices[fragid] = []
            for at in matching[fragid]:
                frag_indices[fragid] += self.atoms[at]["indices"]

        return frag_indices


def compute_spillage_values(sinvxh, sinvxh2, frag_indices, target):
    """
    Computes the actual spillage values.

    Args:
      sinvxh (scipy.sparse.csc): S^-1 * H
      sinvxh2 (scipy.sparse.csc): (S^-1 * H)^2
      frag_indices (dict): list of indices associated with each fragment.
      target (int): which fragment is the target fragment.

    Returns:
      (tuple): ``H2F``,``HGHF``, where the first contribution is the target
      value that the sum of the other contribution should reach.
      Such second term is the interaction spillage between the target and
      each fragment.
    """
    from numpy import trace
    indices_f = frag_indices[target]
    spillage_values = {}

    # Compute the denominator tr(HRfHRf)
    denom = sinvxh[:, indices_f]
    denom = denom[indices_f, :]
    denom = denom.dot(denom)
    denom_t = trace(denom.todense())

    # Compute the left side tr(HS-1HRf)
    H2T = sinvxh2[:, indices_f]
    H2T = H2T[indices_f, :]
    left_t = trace(H2T.todense())

    # Compute the right side values tr(HRgHRf)
    for id_g in frag_indices:
        indices_g = frag_indices[id_g]
        TFH = sinvxh[indices_f, :]
        TFHTG = TFH[:, indices_g]

        TGH = sinvxh[indices_g, :]
        TGHTF = TGH[:, indices_f]

        right_mat = (TFHTG.dot(TGHTF))
        right_t = trace(right_mat.todense())
        spillage_values[id_g] = right_t / denom_t

    return left_t / denom_t, spillage_values


def serial_compute_spillbase(sfile, hfile):
    """
    This routine computes the matrix (S^-1 * H)^2 using python.

    You will need this for computing the spillage values.

    Args:
      sfile (str): the file name of the overlap matrix. Must be in ccs format
      hfile (str): the file name of the hamiltonian. Must be in ccs format

     Returns:
       (scipy.sparse.csc, scipy.sparse.csc): (S^-1 * H), (S^-1 * H)^2
    """
    from scipy.sparse.linalg import inv

    # Read from file
    smat = _read_ccs(sfile)
    hmat = _read_ccs(hfile)

    # Compute the matrix (S^-1H)^2
    sinv = inv(smat)
    sinvxh = sinv.dot(hmat)
    sinvxh2 = sinvxh.dot(sinvxh)

    return sinvxh, sinvxh2


def serial_compute_puritybase(sfile, dfile):
    """
    This routine computes the matrix K*S using python.

    You will need this for computing the purity values.

    Args:
     sfile (str): the file name of the overlap matrix. Must be in ccs format
     dfile (str): the file name of the density kernel. Must be in ccs format

    Returns:
      (scipy.sparse.csc, scipy.sparse.csc): K*S
    """
    # Read from file
    smat = _read_ccs(sfile)
    dmat = _read_ccs(dfile)

    return dmat.dot(smat)


def _read_ccs(fname):
    """
    Read a CCS matrix from file into a scipy csc matrix.

    This routine uses the CCS matrix files that CheSS generates.
    Args:
      fname (str): name of the file.

    Result:
      scipy.sparse.csc_matrix: the matrix in the file.
    """
    from scipy.sparse import csc_matrix

    data = []
    indices = []
    indptr = []

    with open(fname, "r") as ifile:
        # Read the meta data
        line = next(ifile)
        split = line.split()
        matdim = int(split[0])
        nel = int(split[2])
        split = next(ifile).split()
        indptr = [int(x) - 1 for x in split]

        # Read the indices
        added = 0
        while(added < nel):
            split = next(ifile).split()
            values = [int(x) - 1 for x in split]
            indices.extend(values)
            added += len(values)

        # Read the data
        for line in ifile:
            data.append(float(line))

    matrix = csc_matrix((data, indices, indptr), shape=(matdim, matdim))
    return matrix



class Molecule(dict):
    """Class to load a molecule from the BigDFT database"""

    def __init__(self, name):
        import os
        from BigDFT.Fragments import System, Fragment
        from BigDFT.XYZ import XYZReader
        # import the positions of the molecules from the XYZ directory
        dirXYZ = os.path.join(os.path.dirname(__file__), 'XYZs')
        filename = os.path.abspath(os.path.join(dirXYZ, name+'.xyz'))
        if not os.path.isfile(filename):
            raise ValueError('Molecule not available')
        ff = XYZReader(filename)
        frag = Fragment(xyzfile=ff)
        system = System({'molecule:0': frag})
        self.update(system.get_posinp())
        # temporary change of the keys 'values' into 'positions'
        if 'values' in self:
            self['positions'] = self.pop('values')
        if 'positions'in self:
            for at in self['positions']:
                if 'r' in at:
                    at.pop('r')



"""
This module defines some tools used for reading/writing XYZ files.

"""
from futile.Utils import write as safe_print


class XYZReader():
    """
    A class which can be used to read from xyz files.

    This class should behave like a standard file, which means you can use
    it in ``with`` statements, and use the ``next`` command to iterate over it.

    Attributes:
      closed (bool): check if a file is open or not.
      units (str): the units that the xyz file was in.
      natoms (int): the number of atoms in the xyz file.
      cell (list): a list of floats describing the cell.

    Args:
      filename (str): the file to read from. You can also specify a molecule
        that might be in the database.
    """

    def __init__(self, filename):
        self.filename = filename
        self.natoms = None
        self.cell = None
        self.atoms_positions = []

    def open(self):
        self.__enter__()

    def __enter__(self):
        from os.path import join, abspath, dirname
        try:
            self._handle = open(self.filename, "r")
        except IOError:  # see if the file is in the database
            dirXYZ = join(dirname(__file__), "Database", "XYZs")
            filename = abspath(join(dirXYZ, self.filename + ".xyz"))
            self._handle = open(filename, "r")
        line = next(self._handle)
        self.natoms = int(line.split()[0])
        try:
            self.units = line.split()[1]
        except IndexError:
            self.units = "angstroem"
        line = next(self._handle).split()
        if len(line) == 0:
            self.cell = None
        elif line[0] == "#":  # Comment line
            self.cell = None
        elif line[0] == "free":
            self.cell = None
        else:
            self.cell = [float(x) for x in line[1:]]

        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        line = next(self._handle)
        if self.natoms == len(self.atoms_positions):
            raise StopIteration
        split = line.split()
        sym = split[0]
        position = split[1:4]
        this_pos = {'sym': sym, 'r': position, "units": self.units}
        self.atoms_positions.append(this_pos)
        return this_pos

    def __iter__(self):
        return self

    def close(self):
        if self.natoms != len(self.atoms_positions):
            raise IOError('The number of atoms is not consistent with the'
                          ' number of lines')
        self.__exit__()

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        self._handle.close()

    @property
    def closed(self):
        return self._handle.closed


class XYZWriter():
    """
    A class for writing XYZ files.

    This class should behave like a standard file, which means you can use
    it in ``with`` statements and write.

    Args:
      filename (str): the file to write to.
      natoms (int): how many atoms we will write.
      units (str): the units of the file. Defaults to angstroem.
      cell (list): the unit cell.
    """

    def __init__(self, filename, natoms, units="angstroem", cell=None):
        self.filename = filename
        self.natoms = natoms
        self.units = units
        self.cell = cell

    def open(self):
        self.__enter__()

    def __enter__(self):
        self._handle = open(self.filename, "w")
        self._handle.write(str(self.natoms) + " ")
        self._handle.write(self.units)
        self._handle.write("\n")

        # The unit cell
        self._handle.write(_xyz_bc_spec(self.cell))
        self._handle.write("\n")

        return self

    def write(self, atomdict):
        """
        Write an atom to the file.

        Args:
          atom (dict): a dictionary describing the atom.
        """
        from BigDFT.Atom import Atom
        at = Atom(atomdict)
        self._handle.write(at.sym + " ")
        pos = at.get_position(self.units)
        self._handle.write(" ".join([str(x) for x in pos]))
        self._handle.write("\n")

    def close(self):
        self.__exit__()

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        self._handle.close()

    @property
    def closed(self):
        return self._handle.closed


def _xyz_bc_spec(cell):
    """
    Defines the specification for expressing the Boundary Conditions starting
    from a cell vector.

    Args:
      cell (list): array of the (orthorhombic) cell. Should be 0.0 on
        directions with free BC. If None is given, the BC are assumed to
        be Free.

    Return:
     (str): Line of the xyz file specifying the bc
    """
    if cell is None:
        return "free"
    elif cell[1] == 0.0 and cell[2] != 0.0:
        return "surface " + str(cell[0]) + " 0.0 " + str(cell[2]) + " "
    elif cell[1] == 0.0 and cell[2] == 0.0:
        return "wire 0.0 0.0 " + cell[2] + " "
    else:
        return "periodic " + str(cell[0]) + " " + str(cell[1]) + \
               " " + str(cell[2]) + " "


if __name__ == "__main__":
    """Test the XYZ Module"""
    from os.path import join
    from os import system
    for file in ["Si4.xyz", "SiO.xyz"]:
        safe_print("First let's try reading an XYZ file.")
        atom_list = []
        with XYZReader(join("Database", "XYZs", file)) as reader:
            safe_print(reader.closed)
            for at in reader:
                atom_list.append(at)
        safe_print(reader.closed)
        safe_print(atom_list)
        safe_print()

        safe_print("Now let's try writing an XYZ file.")
        safe_print()
        with XYZWriter("test.xyz", len(atom_list), cell=reader.cell,
                       units=reader.units) as writer:
            safe_print(writer.closed)
            for at in atom_list:
                writer.write(at)
        safe_print(writer.closed)
        system("cat test.xyz")
        safe_print()

    safe_print("Using the explicit open and close.")
    reader = XYZReader("test.xyz")
    reader.open()
    print(next(reader))
    reader.close()



from futile.Utils import write as safe_print
from . import InputGenerator

cpickle_file = "bigdft-results.cpickle"


def process_element(run, var, dico, force_run, Elements, strains):
    import math
    import cPickle
    name = dico["name"]
    hgrids = var['dft']['hgrids']
    ngrids = (math.ceil(strains[0] * dico["a"] / hgrids[0]),
              math.ceil(strains[0] * dico["b"] / hgrids[1]),
              math.ceil(strains[0] * dico["c"] / hgrids[2]))
    for strain in strains:
        if run.iproc == 0:
            safe_print("(%s)" % strain)
        if strain in dico["eKS"] and not force_run:
            # The calculation is already done
            if run.iproc == 0:
                safe_print((name, strain), ": ", dico["eKS"][strain])
            continue
        var.update(InputGenerator.set_strain(strain, ngrids, dico))

        run.update(var)

        # Run and store the results
        out = run.run()

        var.update(InputGenerator.set_restart())

        run.update(var)
        # And restart
        out = run.run()

        Elements[name] = dico
        dico["eKS"][strain] = out.eKS
        dico["FKS"][strain] = out.etot
        dico["pressure"][strain] = out.pressure
        if run.iproc == 0:
            cPickle.dump(Elements, open(cpickle_file, "w"))
        # Parse the generated log to get the number of grid points.
        for l in open("log-%s.yaml" % var["radical"], "r"):
            if "Grid Spacing Units" in l:
                grid = l.split()
        ngrids = (int(grid[5][:-1]), int(grid[6][:-1]), int(grid[7]))
        ngrids = [a + 1 for a in ngrids]
    # Freeing memory
    run = None
    out = None
    # Build the file
    if run.iproc == 0:
        fd = open("%s.dat" % name, 'w')
        for strain in strains:
            dico = Elements[name]
            # Volume in A^3/atom
            volume = dico["volume"]*strain**3/dico["nat"]
            HatoeV = 27.21138386
            eKS = float(dico["eKS"][strain])*HatoeV/dico["nat"]
            fd.write("%16.9f %16.9f\n" % (volume, eKS))
            safe_print(strain, dico["eKS"][strain], volume, eKS)
        fd.close()


# Build a dictionary


def elements_from_cif(files, name):
    # filename match as Unix shell
    import fnmatch
    import os
    Elements = dict()
    nonortho = list()
    ortho = list()
    for file in files:
        if not fnmatch.fnmatch(file, "*.cif"):
            continue
        dico = dict()
        dico["file"] = file
        # open("%s/%s" %(dirCIF,file)).readlines():
        for line in open(os.path.abspath(file)).readlines():
            if "_" in line:
                # Remove quote '
                items = line.replace("'", "").split()
                if len(items) == 2:
                    dico[items[0]] = items[1]
            elif "Biso" in line or name in line:
                items = line.split()
                # We know that the maximum of atoms is 8 in the cell
                n = int(items[0][-1])
                dico['nat'] = n
                # dico[n] = "%s  %s  %s " % (items[2], items[3], items[4])
                dico[n] = list(map(float, items[2:5]))
        # Convert name
        if name is not None:
            dico["name"] = name
        else:
            dico["name"] = dico["_pd_phase_name"]
        # We use bohrs
        atob = 1.0/0.5291772108
        dico["a"] = float(dico["_cell_length_a"])*atob
        dico["b"] = float(dico["_cell_length_b"])*atob
        dico["c"] = float(dico["_cell_length_c"])*atob
        dico["alpha"] = float(dico["_cell_angle_alpha"])
        dico["beta"] = float(dico["_cell_angle_beta"])
        dico["gamma"] = float(dico["_cell_angle_gamma"])
        # Only for non-orthorhombic and in angstroem^3
        dico["volume"] = dico["a"]*dico["b"]*dico["c"]/atob**3
        # Create a key results
        dico["eKS"] = dict()
        dico["FKS"] = dict()
        dico["pressure"] = dict()

        InputGenerator.transform_to_orthorombic(dico)

        name = dico['name']
        # Update volume after orthomrombic tranformation and in angstroem^3
        dico["volume"] = dico["a"]*dico["b"]*dico["c"]/atob**3
        if (dico["alpha"] != 90 or dico["beta"] != 90 or
                dico["gamma"] != 90):
            nonortho.append(name)
        else:
            ortho.append(name)
        # Add in the large dictionary
        Elements[name] = dico
    ortho.sort()
    nonortho.sort()
    return Elements, ortho, nonortho

def xyz_from_elements(dirXYZ, Elements, ortho, nonortho):
    import shutil
    import os
    import time
    format_xyz = """{0[nat]} reduced
 periodic {0[a]}   {0[b]}   {0[c]} 
"""
    # print "Remove the directory '%s'." % dirXYZ
    shutil.rmtree(dirXYZ, ignore_errors=True)
    # Create it
    os.mkdir(dirXYZ)
    safe_print("---")
    # Start and create the xyz files
    safe_print("Delta-test timestamp:", time.strftime('%X %x %Z'))
    safe_print("Delta-test code: BigDFT")
    safe_print("Number of elements: ", len(Elements))
    safe_print("List of elements:", list(Elements.keys()))
    safe_print("Number of orthorhombic elements: ", len(ortho))
    safe_print("Orthorhombic elements: ", ortho)
    safe_print("Number of non-orthorhombic elements: ", len(nonortho))
    safe_print("Non-orthorhombic elements: ", nonortho)
    for dico in Elements.values():
        name = dico['name']
        # We have all the specification
        fnew = os.path.join(dirXYZ, name+".xyz")  # "%s/%s.xyz" % (dirXYZ,name)
        fd = open(fnew, "w")
        fd.write(format_xyz.format(dico))
        for i in range(dico['nat']):
            fd.write("%s " % name)
            print(i, dico['nat'])
            fd.write("%f %f %f\n" % tuple(dico[i+1]))
            # fd.write("%s %s\n" % (name,dico[i+1]))
        fd.close()
        # print "#Creation of the file '{0:s}' from '{1:s}/{2:s}'".format(
        #       fnew,dirCIF,dico['file'])

def xyz_from_pymatgendict(dirXYZ, name, dico):
    import shutil
    import os
    import time
    format_xyz = """{1} reduced
 periodic {0[a]}   {0[b]}   {0[c]} 
"""
    #danger zone.. don't do ?
    if dirXYZ != ".":
        print ("Remove the directory '%s'." % dirXYZ)
        shutil.rmtree(dirXYZ, ignore_errors=True)
    # Create it
    os.mkdir(dirXYZ)
    safe_print("---")
    # Start and create the xyz file
    # We have all the specification
    fnew = os.path.join(dirXYZ, name+".xyz")  # "%s/%s.xyz" % (dirXYZ,name)
    fd = open(fnew, "w")
    nat = len(dico['sites'])
    fd.write(format_xyz.format(dico['lattice'],nat))
    for atom in dico['sites']:
        fd.write("%s " % atom['label'])
        fd.write("%f %f %f\n" % tuple(atom['xyz']))
    fd.close()
        # print "#Creation of the file '{0:s}' from '{1:s}/{2:s}'".format(
        #       fnew,dirCIF,dico['file'])

class Benchmark():
    def __init__(self, calculator):
        import os
        # import the positions of the periodic table from the CIFs directory
        dirCIF = os.path.join(os.path.dirname(__file__), 'CIFs')
        files = [os.path.abspath(os.path.join(dirCIF, f))
                 for f in os.listdir(dirCIF)]
        self.Elements, self.ortho, self.nonortho = elements_from_cif(files)
        # Strains (The original)
        self.strains = [0.94, 0.96, 0.98, 1.00, 1.02, 1.04, 1.06]
        # From Francois Jollet
        # strains = [ 0.9795861087155615, 0.986484829732188,
        #             0.9932883883792687,  1.00, 1.006622709560113,
        #             1.0131594038201772, 1.0196128224222163 ]
        # ortho = [ "Ag"]
        self.strains.reverse()
        # strains = [ 1.00 ]
        self.calculator = calculator
        # Now build "*.xyz" for each elements
        if self.calculator.iproc == 0 and False:
            xyz_from_elements("XYZs", self.Elements, self.ortho, self.nonortho)

    def run(self, atoms=None, force=False):
        import os
        # To dump python object in a file
        import cPickle
        if atoms is None:
            atomlist = self.ortho
        else:
            atomlist = atoms
        # Loop over the elements
        # Test if the file bigdft-results.cpickle exists and use it instead
        if os.path.exists(cpickle_file) and not force:
            self.Elements = cPickle.load(open(cpickle_file, "r"))
        for name in atomlist:
            dico = self.Elements[name]
            if self.calculator.iproc == 0:
                safe_print("Start calculation of %s" % dico["name"])
            hgrid = 0.30
            var = InputGenerator.set_inputfile(hgrid, dico)
            self.calculator.set(var)
            process_element(self.calculator, var, dico, force,
                            self.Elements, self.strains)
